import typing

import numpy as _np  # NOQA: ICN001
from numpy.typing import NDArray

from fastfem.fields.field import Field, ShapeComponent

ShapeEntryType = tuple[int, ...]


def assemble_field_add(
    disassembled_field: Field,
    node_indices: NDArray[_np.intp],
    assembled_basis_shape: int | ShapeEntryType,
) -> Field:
    """This function is used for mesh assemblies, where fields on subelements must
    be accumulated into the larger mesh element. For integration, the same node may
    need to take the value of a sum of integrals on subelements. This requires
    an atomic add operation.

    Given an assembly with a 1-axis basis shape B1, with a stack B2 of subelements
    with shape B3, the node index `i` in B1 corresponds to the index K of subelement J
    for which `node_indices[*J,*K] == i`.

    Args:
        disassembled_field (Field): The field with basis shape B3 corresponding to the
            stack of disassembled fields. The leading axes in the stack must match B2.
        node_indices (NDArray[np.intp]): The array of indices demonstrating the
            assembly.
        assembled_basis_shape (int | tuple[int, ...]): The shape of the assembled field.
            This should be 1D.

    Returns:
        Field: The assembled field.
    """
    # verify assembly shapes are compatible
    subelem_basis_shape = disassembled_field.basis_shape
    if node_indices.shape[-len(subelem_basis_shape) :] != subelem_basis_shape:
        message = (
            "The last axes of node_indices must match the basis shape!"
            f" ({node_indices.shape[-len(subelem_basis_shape):]} != basis shape"
            f" {subelem_basis_shape}; node_indices.shape = {node_indices.shape})"
        )
        raise ValueError(message)
    assembly_stack_shape = node_indices.shape[: -len(subelem_basis_shape)]
    if (
        assembly_stack_shape
        != disassembled_field.stack_shape[: len(assembly_stack_shape)]
    ):
        message = (
            "The first axes of node_indices must match the first axes of the stack"
            f" shape! ({node_indices.shape[:-len(subelem_basis_shape)]} !="
            f" {disassembled_field.stack_shape[:len(assembly_stack_shape)]};"
            f" node_indices.shape = {node_indices.shape}; stack_shape ="
            f" {disassembled_field.stack_shape})"
        )
        raise ValueError(message)
    if isinstance(assembled_basis_shape, int):
        assembled_basis_shape = (assembled_basis_shape,)
    if len(assembled_basis_shape) != 1:
        message = (
            "assembled_basis_shape must have 1-dimension! (assembled_basis_shape ="
            f" {assembled_basis_shape})"
        )
        raise ValueError(message)

    if disassembled_field.use_jax:
        message = "JAX-friendly accumulation methods not yet implemented!"
        raise NotImplementedError(message)

    outshape = [()] * 3
    outshape[ShapeComponent.BASIS] = assembled_basis_shape  # type: ignore
    outshape[ShapeComponent.STACK] = disassembled_field.stack_shape[  # type: ignore
        len(assembly_stack_shape) :
    ]
    outshape[ShapeComponent.POINT] = disassembled_field.point_shape  # type: ignore
    accum = _np.zeros(
        outshape[disassembled_field.shape_order_inverse[0]]
        + outshape[disassembled_field.shape_order_inverse[1]]
        + outshape[disassembled_field.shape_order_inverse[2]],
        dtype=disassembled_field.coefficients.dtype,
    )
    src_basis_ind = disassembled_field._axis_field_to_numpy((ShapeComponent.BASIS, 0))
    src_stack_ind = disassembled_field._axis_field_to_numpy((ShapeComponent.STACK, 0))
    end_basis_ind = src_basis_ind - (
        len(assembly_stack_shape)
        if disassembled_field.shape_order[ShapeComponent.STACK]
        < disassembled_field.shape_order[ShapeComponent.BASIS]
        else 0
    )

    src_shift_origin = (
        *range(src_stack_ind, src_stack_ind + len(assembly_stack_shape)),
        *range(src_basis_ind, src_basis_ind + len(subelem_basis_shape)),
    )
    src_shift_target = (
        *range(len(assembly_stack_shape)),
        *range(
            len(assembly_stack_shape),
            len(assembly_stack_shape) + len(subelem_basis_shape),
        ),
    )
    _np.add.at(
        _np.moveaxis(accum, end_basis_ind, 0),
        node_indices,
        _np.moveaxis(
            typing.cast(NDArray, disassembled_field.coefficients),
            src_shift_origin,
            src_shift_target,
        ),
    )

    return Field(
        outshape[ShapeComponent.BASIS],
        outshape[ShapeComponent.POINT],
        accum,
        shape_order=disassembled_field.shape_order,
        use_jax=False,
    )



import itertools
from collections.abc import Sequence
from typing import Literal, cast

import numpy  # NOQA: ICN001
from numpy.typing import ArrayLike

from fastfem.fields.field import (
    Field,
    FieldAxisIndex,
    FieldAxisIndexType,
    ShapeComponent,
    _reshape,
)
from fastfem.fields.field import np_or_jnp as np

from . import numpy_similes_linalg as linalg


def moveaxis(
    field: Field, source: FieldAxisIndexType, destination: FieldAxisIndexType
) -> Field:
    """This attempts to replicate the numpy `moveaxis` function. Currently, multiple axes at the same time are not supported.
    https://numpy.org/doc/stable/reference/generated/numpy.moveaxis.html

    Args:
        field (Field): The field whose axes should be reordered
        source (FieldAxisIndexType | typing.Sequence[FieldAxisIndexType]): Original positions of the axes to move. These must be unique.
        destination (FieldAxisIndexType | typing.Sequence[FieldAxisIndexType]): Destination positions of the axes to move. These must also be unique.
    """

    shapes = {
        ShapeComponent.BASIS: field.basis_shape,
        ShapeComponent.STACK: field.stack_shape,
        ShapeComponent.POINT: field.point_shape,
    }

    src_pos_np = field._axis_field_to_numpy(source)
    srcshape = shapes[source[0]]

    src_pt = source[1] if source[1] >= 0 else len(srcshape) + source[1]
    rem_axis = srcshape[source[1]]  # size of the removed axis
    # when counting from the right, be sure to add one for the new length
    dest_pt = (
        destination[1]
        if destination[1] >= 0
        else len(shapes[destination[0]])
        + (0 if source[0] == destination[0] else 1)
        + destination[1]
    )
    dest_pos_np = field._axis_field_to_numpy((destination[0], dest_pt), False)
    coefs = np(field.use_jax).moveaxis(
        field.coefficients,
        src_pos_np,
        # dest_pos_np - (1 if dest_pos_np > src_pos_np else 0), #account for removals shifting indices
        dest_pos_np
        - (
            1 if field.shape_order[source[0]] < field.shape_order[destination[0]] else 0
        ),
    )
    shapes[source[0]] = srcshape[:src_pt] + srcshape[(src_pt + 1) :]
    destshape = shapes[destination[0]]
    shapes[destination[0]] = destshape[:dest_pt] + (rem_axis,) + destshape[dest_pt:]
    return Field(
        shapes[ShapeComponent.BASIS],
        shapes[ShapeComponent.POINT],
        coefs,
        shape_order=field.shape_order,
    )


def sum(
    field: Field,
    axes: FieldAxisIndexType | tuple[FieldAxisIndexType, ...] | ShapeComponent | None,
) -> Field:
    """This attempts to replicate the numpy "sum" function. In
    https://numpy.org/doc/stable/reference/generated/numpy.sum.html

    Args:
        field (Field): _description_
        axes (FieldAxisIndex | tuple[FieldAxisIndex,...] | ShapeComponent | None): _description_

    Returns:
        Field: _description_
    """
    if axes is None:
        return Field((), (), np(field.use_jax).sum(field.coefficients))
    axes_: tuple[FieldAxisIndexType, ...] = ()
    if isinstance(axes, ShapeComponent):
        axes_ = tuple(
            FieldAxisIndex(axes, i) for i in range(len(field.get_shape(axes)))
        )
    elif isinstance(axes, FieldAxisIndex):
        axes_ = (axes,)
    elif isinstance(axes[0], tuple | FieldAxisIndex):
        # ^ force tuple[tuple]              ^ force tuple[FieldAxisIndex]

        # for some reason pyright doesn't approve
        axes_ = cast(tuple[FieldAxisIndexType, ...], axes)
    else:
        axes_ = cast(tuple[FieldAxisIndexType, ...], (axes,))

    coefs = np(field.use_jax).sum(
        field.coefficients, tuple(map(field._axis_field_to_numpy, axes_))
    )
    shapes = {
        ShapeComponent.BASIS: field.basis_shape,
        ShapeComponent.STACK: field.stack_shape,
        ShapeComponent.POINT: field.point_shape,
    }
    for ax in axes_:
        shape = shapes[ax[0]]
        shapes[ax[0]] = shape[: ax[1]] + shape[(ax[1] + 1) :]
    return Field(shapes[ShapeComponent.BASIS], shapes[ShapeComponent.POINT], coefs)


def reshape(
    field: Field,
    component_selector: ShapeComponent,
    shape: int | tuple[int],
    order: Literal["C", "F", "A"] = "C",
    copy: bool | None = None,
) -> Field:
    """This attempts to replicate the numpy "reshape" function.
    https://numpy.org/doc/stable/reference/generated/numpy.reshape.html

    reshape() applies numpy "reshape" to a given component. This function is also called
    when using the field accessor reshape methods. That is, `reshape(field,BASIS,s)` is
    the same as `field.basis.reshape(s)`
    Args:
        field (Field): The field to reshape
        component_selector (ShapeComponent): Which component to reshape.
        shape (int | tuple[int]): The target shape of the component. For any integer i,
            i is equivalent to (i,)
        order ({'C','F','A'}, optional): See the numpy documentation. Defaults to 'C'.
        copy (bool | None, optional): See the numpy documentation. Defaults to None.

    Raises:
        ValueError: when a copy operation is required, but `copy` is False.
    Returns:
        Field: The reshaped field. This is always a new object, but if data is copied,
        the underlying array is a view of the original.
    """
    return _reshape(field, component_selector, shape, order, copy)


def einsum(
    component_selector: ShapeComponent,
    subscripts: str,
    *operands: ArrayLike | Field,
) -> Field:
    fieldops = tuple(op for op in operands if isinstance(op, Field))
    if len(fieldops) == 0:
        message = "Please provide at least one Field!"
        raise ValueError(message)

    valid_indices = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    desired_order = fieldops[0].shape_order

    # which components to reshape to single lines; we need indices for them:
    comp1rs = ShapeComponent(1 if component_selector == 0 else 0)
    comp2rs = ShapeComponent(
        (comp1rs + 2) if component_selector == (comp1rs + 1) else comp1rs + 1
    )
    # order so that the first one appears first
    if desired_order[comp1rs] > desired_order[comp2rs]:
        comp1rs, comp2rs = comp2rs, comp1rs

    # broadcast all components except the one being selected
    broadcast_shape = [()] * 3
    for comp in comp1rs, comp2rs:
        broadcast_shape[comp] = numpy.broadcast_shapes(  # type: ignore
            *(field.get_shape(comp) for field in fieldops)
        )
    broadcasted = []
    for field in fieldops:
        broadcast_shape[component_selector] = field.get_shape(component_selector)  # type: ignore
        broadcasted.append(
            field.broadcast_to_shape(
                broadcast_shape[ShapeComponent.STACK],
                broadcast_shape[ShapeComponent.BASIS],
                broadcast_shape[ShapeComponent.POINT],
            )
        )

    if desired_order[comp1rs] < desired_order[component_selector]:
        comp1label = valid_indices[0]
        if desired_order[comp2rs] < desired_order[component_selector]:
            comp2label = valid_indices[1]
        else:
            comp2label = valid_indices[-1]
    else:
        comp1label = valid_indices[-2]
        comp2label = valid_indices[-1]

    if comp1label in subscripts or comp2label in subscripts:
        message = (
            f"Subscripts {comp1label} and {comp2label} have not been reserved for"
            " non-selected indices!"
        )
        raise ValueError(message)

    # the shapes of the result
    output_shapes: list[tuple[int, ...]] = [()] * 3
    output_shapes[desired_order[comp1rs]] = broadcasted[0].get_shape(comp1rs)
    output_shapes[desired_order[comp2rs]] = broadcasted[0].get_shape(comp2rs)
    use_jax = any(f.use_jax for f in broadcasted)
    # inject broadcasted operands back
    broadcasted_iter = iter(broadcasted)
    operands = tuple(
        (
            reshape(
                reshape(
                    next(broadcasted_iter),
                    comp1rs,
                    -1,
                ),
                comp2rs,
                -1,
            )
            if isinstance(op, Field)
            else op
        )
        for op in operands
    )

    def component_inject(order) -> tuple[str, str]:
        compstr = (
            comp1label + comp2label
            if order[comp1rs] < order[comp2rs]
            else comp2label + comp1label
        )
        ind = order[component_selector]
        return compstr[:ind], compstr[ind:]

    ops_iter = iter(
        itertools.chain(
            (
                (component_inject(op.shape_order) if isinstance(op, Field) else None)
                for op in operands
            ),
            [component_inject(desired_order)],
        )
    )
    einstr = "->".join(
        ",".join(
            (subs if ins is None else ins[0] + subs + ins[1])
            for subs, ins in zip(sec.split(","), ops_iter, strict=False)
        )
        for sec in subscripts.split("->")
    )
    coefs = np(*broadcasted).einsum(
        einstr,
        *(op.coefficients if isinstance(op, Field) else op for op in operands),
    )
    if desired_order[component_selector] == 0:
        output_shapes[desired_order[component_selector]] = coefs.shape[:-2]
    elif desired_order[component_selector] == 1:
        output_shapes[desired_order[component_selector]] = coefs.shape[1:-1]
    else:
        output_shapes[desired_order[component_selector]] = coefs.shape[2:]
    return Field(
        output_shapes[desired_order[ShapeComponent.BASIS]],
        output_shapes[desired_order[ShapeComponent.POINT]],
        coefs.reshape(
            output_shapes[0] + output_shapes[1] + output_shapes[2],
        ),
        shape_order=broadcasted[0].shape_order,
        use_jax=use_jax,
    )


def abs(field: Field) -> Field:
    """absolute value of field. TODO link to numpy

    Args:
        field (Field): _description_

    Returns:
        Field: _description_
    """
    return Field(
        field.basis_shape,
        field.point_shape,
        np(field).abs(field.coefficients),
        shape_order=field.shape_order,
        use_jax=field.use_jax,
    )


def stack(fields: Sequence[Field], axis: FieldAxisIndexType) -> Field:
    """stack multiple fields. TODO link to numpy

    Args:
        fields (Sequence[Field]): _description_
        axis (FieldAxisIndexType): _description_

    Returns:
        Field: _description_
    """
    _np = np(*fields)
    shapes = [()] * 3
    shapes[ShapeComponent.BASIS] = fields[0].basis_shape  # type: ignore
    shapes[ShapeComponent.STACK] = fields[0].stack_shape  # type: ignore
    shapes[ShapeComponent.POINT] = fields[0].point_shape  # type: ignore
    shapes[axis[0]] = (  # type: ignore
        shapes[axis[0]][: axis[1]] + (len(fields),) + shapes[axis[0]][axis[1] :]
    )
    return Field(
        shapes[ShapeComponent.BASIS],
        shapes[ShapeComponent.POINT],
        _np.stack(
            [field.coefficients for field in fields],
            axis=fields[0]._axis_field_to_numpy(axis, False),
        ),
        shape_order=fields[0].shape_order,
        use_jax=fields[0].use_jax,
    )


__all__ = [
    "abs",
    "einsum",
    "linalg",
    "moveaxis",
    "reshape",
    "sum",
]



import numpy as np

import fastfem.fields.numpy_similes as fnp
from fastfem.elements.element2d import Element2D
from fastfem.fields.field import Field, ShapeComponent


class LinearSimplex2D(Element2D):
    """A triangular isparametric element with a 3-dimensional shape function space -
    one for each vertex.
    """

    def basis_shape(self):
        return (3,)

    def reference_element_position_field(self):
        return Field((3,), (2,), np.array([[0, 0], [1, 0], [0, 1]]))

    def _interpolate_field(self, field, X, Y):
        X = Field(X.shape, (), X)
        Y = Field(Y.shape, (), Y)
        return (
            field.basis[0, ...] * (1.0 - X - Y)
            + field.basis[1, ...] * X
            + field.basis[2, ...] * Y
        )

    def _compute_field_gradient(self, field, pos_field=None):
        if (
            field.basis_shape == ()
            or len(field.coefficients.shape) == 0
            or field.coefficients.shape[0] == 1
        ):  # we have a constant function.
            return Field(
                (),
                (
                    *field.point_shape,
                    2,
                ),
                np.zeros(field.stack_shape + field.point_shape + (1,)),
            )

        grad_coefs = fnp.stack(
            [
                field.basis[1, ...] - field.basis[0, ...],
                field.basis[2, ...] - field.basis[0, ...],
            ],
            axis=(ShapeComponent.POINT, len(field.point_shape)),
        )
        if pos_field is not None:
            def_grad = self._compute_field_gradient(pos_field)
            grad_coefs = fnp.einsum(
                ShapeComponent.POINT,
                "...ij,...i->...j",
                fnp.linalg.inv(def_grad, ShapeComponent.POINT),
                grad_coefs,
            )

        return grad_coefs

    def _integrate_field(self, pos_field, field, jacobian_scale=...):
        coefs = (np.ones((3, 3)) + np.eye(3)) / 24
        return fnp.einsum(
            ShapeComponent.BASIS,
            ",ij,i,j->",
            fnp.abs(fnp.linalg.det(self._compute_field_gradient(pos_field))),
            coefs,
            field,
            jacobian_scale,
        )

    def _integrate_basis_times_field(
        self, pos_field, field, indices=None, jacobian_scale=...
    ):
        coefs = (
            np.array(
                [
                    [[6, 2, 2], [2, 2, 1], [2, 1, 2]],
                    [[2, 2, 1], [2, 6, 2], [1, 2, 2]],
                    [[2, 1, 2], [1, 2, 2], [2, 2, 6]],
                ]
            )
            / 120
        )
        res = fnp.einsum(
            ShapeComponent.BASIS,
            ",kij,i,j->k",
            fnp.abs(fnp.linalg.det(self._compute_field_gradient(pos_field))),
            coefs,
            field,
            jacobian_scale,
        )
        return res if indices is None else res.basis[*indices]

    def _integrate_grad_basis_dot_field(
        self, pos_field, field, indices=None, jacobian_scale=...
    ):
        # this is rather unoptimized. TODO make better
        basis_diff_coefs = np.array([[-1, -1], [1, 0], [0, 1]])
        defgrad = self._compute_field_gradient(pos_field)
        dginv = fnp.linalg.inv(defgrad)

        # pad to field-shape, excluding last axis, which is dotted (contracted)
        # fieldpad = (np.newaxis,) * (len(field.point_shape) - 1)
        basis_times_field = fnp.einsum(
            ShapeComponent.POINT,
            ",kl,lg,...g->k...",
            # exclude jacobian, since we are delegating to integrate_field subroutine.
            # np.abs(np.linalg.det(defgrad.coefficients)[..., *fieldpad]),
            1,
            basis_diff_coefs,
            dginv,
            field,
        )
        integ = fnp.moveaxis(
            self.integrate_field(
                pos_field,
                basis_times_field,
                jacobian_scale,
            ),
            (ShapeComponent.POINT, 0),
            (ShapeComponent.BASIS, 0),
        )
        if indices is None:
            return integ
        return integ.basis[*indices]



import dataclasses
import pathlib
from typing import Literal, Optional

from .generator import (
    Domain,
    Geometry,
    Line,
    Mesh,
    Point,
    Surface,
    TwoDElementType,
    mesh,
)

__all__ = ["Rectangle", "Square", "create_a_rectangle_mesh", "create_a_square_mesh"]


def __dir__() -> list[str]:
    return __all__


@dataclasses.dataclass(kw_only=True)
class Rectangle:
    """Create a 2D mesh of a rectangle.

    Args:
        horizontal_length (float): The horizontal length of the rectangle.
        vertical_length (float): The vertical length of the rectangle.
        nodes_in_horizontal_direction (int, optional): The number of nodes in the
            horizontal direction. Defaults to None.
        nodes_in_vertical_direction (int, optional): The number of nodes in the vertical
            direction. Defaults to None.
        element_type (TwoDElementType, optional): The type of element to use. Defaults
            to "quadrangle".
    """

    horizontal_length: float
    vertical_length: float
    nodes_in_horizontal_direction: Optional[int] = None
    nodes_in_vertical_direction: Optional[int] = None
    element_type: TwoDElementType = "quadrangle"

    def __post_init__(
        self,
    ):
        both_nx_and_ny_are_provided = all(
            [self.nodes_in_horizontal_direction, self.nodes_in_vertical_direction]
        )
        self.transfinite = False
        if both_nx_and_ny_are_provided:
            self.transfinite = True

        self.surface = Surface(
            outer_lines=[
                Line(
                    Point(0, 0, 0),
                    Point(self.horizontal_length, 0, 0),
                    self.nodes_in_horizontal_direction,
                ),
                Line(
                    Point(self.horizontal_length, 0, 0),
                    Point(self.horizontal_length, self.vertical_length, 0),
                    self.nodes_in_vertical_direction,
                ),
                Line(
                    Point(self.horizontal_length, self.vertical_length, 0),
                    Point(0, self.vertical_length, 0),
                    self.nodes_in_horizontal_direction,
                ),
                Line(
                    Point(0, self.vertical_length, 0),
                    Point(0, 0, 0),
                    self.nodes_in_vertical_direction,
                ),
            ],
            transfinite=self.transfinite,
            element_type=self.element_type,
        )
        Geometry().mesh()


@dataclasses.dataclass(kw_only=True)
class Square(Rectangle):
    """Create a 2D mesh of a square.

    Args:
        side_length (float): The side length of the square.
        nodes_in_horizontal_direction (int, optional): The number of nodes in the
            horizontal direction. Defaults to None.
        nodes_in_vertical_direction (int, optional): The number of nodes in the vertical
            direction. Defaults to None.
        element_type (TwoDElementType, optional): The type of element to use. Defaults
            to "quadrangle".
    """

    side_length: float
    horizontal_length: float = dataclasses.field(init=False)
    vertical_length: float = dataclasses.field(init=False)

    def __post_init__(
        self,
    ):
        self.horizontal_length = self.side_length
        self.vertical_length = self.side_length
        super().__post_init__()


RectangleDomainName = Literal[
    "bottom_boundary",
    "right_boundary",
    "top_boundary",
    "left_boundary",
    "surface",
]


class RectangleMesh(Mesh):
    """A class that is identical to the `Mesh` class, but with typing hints for a
    rectangle mesh."""

    def __getitem__(  # type: ignore
        self,
        key: RectangleDomainName,
    ) -> Domain:
        return self[key]


class SquareMesh(RectangleMesh): ...


def create_a_rectangle_mesh(
    horizontal_length: float,
    vertical_length: float,
    nodes_in_horizontal_direction: Optional[int] = None,
    nodes_in_vertical_direction: Optional[int] = None,
    element_type: TwoDElementType = "quadrangle",
    file_name: Optional[pathlib.Path] = None,
) -> RectangleMesh:
    """Create a 2D mesh of a rectangle.

    Args:
        horizontal_length: The horizontal length of the rectangle.
        vertical_length: The vertical length of the rectangle.
        nodes_in_horizontal_direction: The number of nodes in the horizontal direction.
            Defaults to None.
        nodes_in_vertical_direction: The number of nodes in the vertical direction.
            Defaults to None.
        element_type: The type of element to use. Defaults to "quadrangle".
        file_name: The file name to save the mesh to. Defaults to None.

    Returns:
        The mesh of the rectangle.
    """

    both_nx_and_ny_are_provided = all(
        [nodes_in_horizontal_direction, nodes_in_vertical_direction]
    )
    transfinite = False
    if both_nx_and_ny_are_provided:
        transfinite = True

    Surface(
        outer_lines=[
            Line(
                Point(0, 0, 0),
                Point(horizontal_length, 0, 0),
                nodes_in_horizontal_direction,
                domain_name="bottom_boundary",
            ),
            Line(
                Point(horizontal_length, 0, 0),
                Point(horizontal_length, vertical_length, 0),
                nodes_in_vertical_direction,
                domain_name="right_boundary",
            ),
            Line(
                Point(horizontal_length, vertical_length, 0),
                Point(0, vertical_length, 0),
                nodes_in_horizontal_direction,
                domain_name="top_boundary",
            ),
            Line(
                Point(0, vertical_length, 0),
                Point(0, 0, 0),
                nodes_in_vertical_direction,
                domain_name="left_boundary",
            ),
        ],
        transfinite=transfinite,
        element_type=element_type,
        domain_name="surface",
    )
    return mesh(file_name=file_name)  # type: ignore


def create_a_square_mesh(
    side_length: float,
    nodes_in_horizontal_direction: Optional[int] = None,
    nodes_in_vertical_direction: Optional[int] = None,
    element_type: TwoDElementType = "quadrangle",
    file_name: Optional[pathlib.Path] = None,
) -> SquareMesh:
    """Create a 2D mesh of a square.

    Args:
        side_length: The side length of the square.
        nodes_in_horizontal_direction: The number of nodes in the horizontal direction.
            Defaults to None.
        nodes_in_vertical_direction: The number of nodes in the vertical direction.
            Defaults to None.
        element_type: The type of element to use. Defaults to "quadrangle".
    """
    horizontal_length = side_length
    vertical_length = side_length
    return create_a_rectangle_mesh(
        horizontal_length,
        vertical_length,
        nodes_in_horizontal_direction,
        nodes_in_vertical_direction,
        element_type,
        file_name,
    )  # type: ignore



"""
FastFEM is planned to be a general-purpose finite element method (FEM) library with a
focus on great Python interface. Currently, it can only solve

$$
\\frac{\\partial^2 f(x,y,t)}{\\partial x^2} + \\frac{\\partial^2 f(x,y,t)}{\\partial y^2}
=
h(f) \\frac{\\partial f(x,y,t)}{\\partial t} + g(x,y)
$$

where $f(x,y,t)$, $h(f)$, and $g(x,y)$ are scalar functions, $x$ and $y$ are spatial
coordinates, and $t$ is time.
"""

__version__ = "0.0.1"

from . import mesh

__all__ = ["mesh"]



import gmsh
import pytest

import fastfem.mesh as m
import fastfem.mesh.generator as mg


def setup_function(function):
    """Setup for test functions"""
    mg.Geometry().clear()


def test_points_database():
    p1 = mg.Point(0.0, 0.0, 0.0)
    p2 = mg.Point(0.0, 0.0, 2.0)
    p3 = mg.Point(0.0, 0.0, 0.0)

    assert p1.tag == p3.tag
    assert mg.Geometry().points == [p1, p2]


def test_lines_database():
    p1 = mg.Point(0.0, 0.0, 0.0)
    p2 = mg.Point(1.0, 0.0, 0.0)
    p3 = mg.Point(1.0, 1.0, 0.0)

    l1 = mg.Line(p1, p2)
    l2 = mg.Line(p1, p3)
    l3 = mg.Line(p1, p2)
    l4 = mg.Line(p2, p1)

    assert l1.tag == l3.tag
    assert l4.tag == -l1.tag
    assert mg.Geometry().lines == [l1, l2]


def test_surfaces_database():
    p1 = mg.Point(0.0, 0.0, 0.0)
    p2 = mg.Point(1.0, 0.0, 0.0)
    p3 = mg.Point(1.0, 1.0, 0.0)
    p4 = mg.Point(0.0, 1.0, 0.0)

    l1 = mg.Line(p1, p2)
    l2 = mg.Line(p2, p3)
    l3 = mg.Line(p3, p1)
    l4 = mg.Line(p3, p4)
    l5 = mg.Line(p4, p1)

    s1 = mg.Surface([l1, l2, l3])
    s2 = mg.Surface([l1, l2, l4, l5])
    s3 = mg.Surface([l1, l2, l3])

    assert s1.tag == s3.tag
    assert mg.Geometry().surfaces == [s1, s2]


def test_negative_duplicate_lines():
    l1 = mg.Line(mg.Point(0.0, 0.0, 0.0), mg.Point(1.0, 0.0, 0.0))
    l2 = mg.Line(mg.Point(1.0, 0.0, 0.0), mg.Point(0.0, 0.0, 0.0))

    assert l1.tag == -l2.tag


def test_invalid_surface_mesh():
    l1 = mg.Line(mg.Point(0.0, 0.0, 0.0), mg.Point(1.0, 0.0, 0.0))
    l2 = mg.Line(mg.Point(1.0, 0.0, 0.0), mg.Point(1.0, 1.0, 0.0))
    l3 = mg.Line(mg.Point(1.0, 1.0, 0.0), mg.Point(0.0, 1.0, 0.0))
    l4 = mg.Line(mg.Point(0.0, 1.0, 0.0), mg.Point(0.0, 0.0, 0.0))

    with pytest.raises(ValueError):
        # Surface cannot be transfinite if lines are not transfinite
        # (number_of_nodes is None for at least one line)
        mg.Surface([l1, l2, l3, l4], transfinite=True)


def test_invalid_surface_mesh_2():
    l1 = mg.Line(mg.Point(0.0, 0.0, 0.0), mg.Point(0.5, 0.0, 0.0), number_of_nodes=10)
    l2 = mg.Line(mg.Point(0.5, 0.0, 0.0), mg.Point(1.0, 0.0, 0.0), number_of_nodes=10)
    l3 = mg.Line(mg.Point(1.0, 0.0, 0.0), mg.Point(1.0, 1.0, 0.0), number_of_nodes=10)
    l4 = mg.Line(mg.Point(1.0, 1.0, 0.0), mg.Point(0.0, 1.0, 0.0), number_of_nodes=10)
    l5 = mg.Line(mg.Point(0.0, 1.0, 0.0), mg.Point(0.0, 0.0, 0.0), number_of_nodes=10)

    with pytest.raises(ValueError):
        # Transfinite surfaces must have 3 or 4 lines
        mg.Surface([l1, l2, l3, l4, l5], transfinite=True)


def test_invalid_surface_mesh_3():
    l1 = mg.Line(mg.Point(0.0, 0.0, 0.0), mg.Point(0.5, 0.0, 0.0))
    l2 = mg.Line(mg.Point(0.5, 0.0, 0.0), mg.Point(1.0, 0.0, 0.0))
    with pytest.raises(ValueError):
        mg.Surface([l1, l2])


def test_surface_with_lines_in_wrong_order():
    l1 = mg.Line(mg.Point(0.0, 0.0, 0.0), mg.Point(1.0, 0.0, 0.0))
    l2 = mg.Line(mg.Point(1.0, 0.0, 0.0), mg.Point(1.0, 1.0, 0.0))
    l3 = mg.Line(mg.Point(1.0, 1.0, 0.0), mg.Point(0.0, 1.0, 0.0))
    l4 = mg.Line(mg.Point(0.0, 1.0, 0.0), mg.Point(0.0, 0.0, 0.0))

    with pytest.raises(ValueError):
        mg.Surface([l3, l2, l1, l4])


def test_surface_with_lines_correct_order_but_wrong_orientation():
    l1 = mg.Line(mg.Point(0.0, 0.0, 0.0), mg.Point(1.0, 0.0, 0.0))
    l2 = mg.Line(mg.Point(1.0, 0.0, 0.0), mg.Point(1.0, 1.0, 0.0))
    l3 = mg.Line(mg.Point(1.0, 1.0, 0.0), mg.Point(0.0, 1.0, 0.0))
    l4 = mg.Line(mg.Point(0.0, 1.0, 0.0), mg.Point(0.0, 0.0, 0.0))

    mg.Surface([l1, l2, -l3, l4])


def test_invalid_domains():
    with pytest.raises(ValueError):
        mg.Line(
            mg.Point(0.0, 0.0, 0.0),
            mg.Point(1.0, 0.0, 0.0, domain_name="bottom_boundary"),
            domain_name="bottom_boundary",
        )
        mg.Geometry().create_domains()


def test_valid_domains():
    l1 = mg.Line(
        mg.Point(1.0, 0.0, 0.0),
        mg.Point(1.0, 1.0, 0.0),
        domain_name="right_boundary",
    )
    l2 = mg.Line(
        mg.Point(1.0, 1.0, 0.0),
        mg.Point(1.0, 2.0, 0.0),
        domain_name="right_boundary",
    )

    mg.Geometry().create_domains()

    entities = list(gmsh.model.get_entities_for_physical_group(1, 1))
    entities = [int(tag) for tag in entities]

    assert entities == [l1.tag, l2.tag]


@pytest.mark.parametrize(
    "nx",
    [
        10,
        None,
    ],
)
@pytest.mark.parametrize(
    "ny",
    [
        10,
        None,
    ],
)
@pytest.mark.parametrize(
    "element_type",
    [
        "triangle",
        "quadrangle",
    ],
)
@pytest.mark.parametrize(
    "function",
    [
        lambda nx, ny, type: m.create_a_rectangle_mesh(
            horizontal_length=1.0,
            vertical_length=1.0,
            nodes_in_horizontal_direction=nx,
            nodes_in_vertical_direction=ny,
            element_type=type,
        ),
        lambda nx, ny, type: m.create_a_square_mesh(
            side_length=1.0,
            nodes_in_horizontal_direction=nx,
            nodes_in_vertical_direction=ny,
            element_type=type,
        ),
    ],
)
def test_rectangle_and_square_mesh(nx, ny, element_type, function):
    mesh: m.SquareMesh | m.RectangleMesh = function(
        nx=nx,
        ny=ny,
        type=element_type,
    )
    assert "bottom_boundary" in mesh
    assert "right_boundary" in mesh
    assert "top_boundary" in mesh
    assert "left_boundary" in mesh
    assert "surface" in mesh
    for domain in mesh:
        if domain.name == "surface":
            assert domain.dimension == 2
            if element_type == "triangle":
                for tag, element_nodes in domain.mesh[0].elements.items():
                    assert len(element_nodes) == 3
                    assert isinstance(tag, int)
            else:
                for tag, element_nodes in domain.mesh[0].elements.items():
                    assert len(element_nodes) == 4
                    assert isinstance(tag, int)
        else:
            assert domain.dimension == 1
            for tag, node_coordinates in domain.mesh[0].nodes.items():
                assert len(node_coordinates) == 3
                assert isinstance(tag, int)
    if nx and ny:
        assert mesh.number_of_nodes == nx * ny

    for tag, node_coordinates in mesh.nodes.items():
        assert len(node_coordinates) == 3
        assert isinstance(tag, int)

    assert len(list(mesh)) == 5



import abc
import warnings

import numpy as np
from numpy.typing import ArrayLike

from fastfem.fields.field import Field as FieldType

WARNINGS_ENABLED = False


def _set_warnings_enabled(enable_warnings: bool):
    """Elements may send developer warnings if a default method is used (for example,
    gradient interpolation computes the gradient field, then interpolates it by default,
    which may be slower than computing gradients only at interpolation points). These
    warnings can be enabled/disabled through this function.

    Args:
        enable_warnings (bool): True to enable warnings. False to disable.
    """
    global WARNINGS_ENABLED  # NOQA: PLW0603
    WARNINGS_ENABLED = enable_warnings


class ElementBase(abc.ABC):  # NOQA: B024
    """
    Template to handle management of element operations.
    Element classes define operations to compute integrals on them.
    The elements themselves are handled in a data-driven
    way, where positions and fields are handled as arrays of coefficients of the shape
    functions defined by the element class. Derived classes must create
    the definitions for interpolation/integration, since ElementBase does not assume
    dimension, or special conditions.
    """

    def _dev_warn(self, message: str):
        """To be used by developers, only.

        This method is called by the base `Element` class to warn the use of default
        (most likely unoptimized, or potentially undesired)
        strategies to compute certain values.

        For example, `Element.mass_matrix()` computes the mass matrix by delegating to
        `integrate_basis_times_field(...,field=basis_fields())`, which may not be very
        efficient.

        This method can be overridden by the developer to suppress these warnings,
        and new ones can be called oustide of the `Element` base class without issue.

        Args:
            message (str): _description_

        Returns:
            _type_: _description_
        """
        if WARNINGS_ENABLED:
            warnings.warn(f"Element developer message: {message}", stacklevel=3)


class IsoparametricElement(abc.ABC):
    """This type is for elements that are isoparametric (shape functions and position
    fields are of the same function space).

    From this, we can specify a consistent shape corresponding to the basis of
    this space.
    """

    @abc.abstractmethod
    def basis_shape(self) -> tuple[int, ...]:
        """
        Each element has a basis with a corresponding multi-index. The shape
        corresponding to the multi-index is given by this method. Any function defined
        on elements is passed into `element`'s methods as a tensor `f` of coefficients
        for the basis, the indices of which should be the leading indices of `f.shape`.

        Returns:
            tuple[int, ...]: a tuple representing the shape of the array corresponding
                to the basis coefficients. A scalar field `f`, given as an array is
                expected to have shape `f.shape == element.basis_shape()`.
        """
        raise NotImplementedError

    # this emulates a member class definition
    def Field(
        self, field: ArrayLike, is_const: bool, point_shape: tuple[int, ...] = ()
    ) -> FieldType:
        """Converts a `numpy` array to a `Field` object that can be used by this
        element class.

        Args:
            field (NDArray): An array representing the field.
            is_const (bool): If `True`, the field is considered
                constant, so `field.shape` is the same as its pointwise shape. For
                scalar fields, this means that `field.shape` is the field's stack shape.
                Otherwise, the first `len(element.basis_shape())` axes of `field` are
                taken to be the basis axes.
            point_shape (tuple[int, ...], optional): The shape of the field (pointwise,
            per-element). Defaults to tuple() for a scalar field.

        Raises:
            ValueError: when `is_const` is `False`, and the first
                `len(element.basis_shape())` axes of `field` are incompatible with
                `element.basis_shape()`.

        Returns:
            Field: An object corresponding to the given field.
        """
        basis_shape = self.basis_shape()
        if not isinstance(field, np.ndarray):
            field = np.array(field)
        return FieldType(
            basis_shape,
            point_shape,
            (
                field[*tuple(np.newaxis for _ in range(len(basis_shape))), ...]
                if is_const
                else field
            ),
        )

    def basis_fields(self) -> FieldType:
        """
        Returns a stack of scalar fields representing the basis elements.
        This is the identity matrix within the `basis_shape` indexing. That is,
        for two multi-indices, `I` and `J`,
        `basis_fields.coefficients[*I,*J] = (I == J)`.

        Returns:
            Field: The Field object with a stack of shape `basis_shape()` representing
                each basis function as a scalar field.
        """
        basis_shape = self.basis_shape()
        field = np.zeros(basis_shape + basis_shape)
        basis_size = np.prod(basis_shape, dtype=int)
        enumeration = np.unravel_index(np.arange(basis_size), basis_shape)
        field[*enumeration, *enumeration] = 1
        return FieldType(basis_shape, (), field)

    def _verify_field_compatibilities(
        self, *fields: FieldType, **named_fields: FieldType
    ):
        compare_field = FieldType(self.basis_shape(), (), 0)
        if not FieldType.are_compatible(compare_field, *fields, *named_fields.values()):
            shapestrs = [str(f.shape) for f in fields] + [
                f"{name}: {f.shape!s}" for name, f in named_fields.items()
            ]
            raise ValueError("Incompatible shapes: " + ", ".join(shapestrs))



from collections.abc import Callable

import numpy as np
import pytest

from fastfem.elements import spectral_element2d as spectral_element
from fastfem.fields.field import Field


@pytest.fixture(scope="module", params=[3, 4, 5])
def element(request):
    order = request.param
    elem = spectral_element.SpectralElement2D(order)
    return elem


@pytest.fixture(scope="module")
def transformed_element(element, transformation):
    return (
        element,
        Field(
            element.basis_shape(),
            (2,),
            transformation(element.reference_element_position_field().coefficients),
        ),
        transformation,
    )


@pytest.fixture(scope="module")
def transformed_element_stack(transform_stack, element):
    transformed = transform_stack(element.reference_element_position_field())
    ndims = len(transformed.shape) - len(element.basis_shape())
    pts_stack = np.permute_dims(
        transformed, (ndims, ndims + 1) + tuple(range(ndims)) + (-1,)
    )
    return element, pts_stack, transform_stack


@pytest.fixture(
    params=[(0, 0), (-1, -1), (1, -1), (1, 1), (-1, 1), (0.5, 0.5), (-0.33, 0.84)]
)
def ref_coords(request):
    return request.param


@pytest.fixture(
    params=[
        np.array((0, 0)),
        np.array(((-1, -1), (1, -1), (1, 1), (-1, 1))),
        np.array((((-0.5, -0.3), (0.7, -0.2)), ((0.2, 0.8), (-0.1, 1)))),
    ]
)
def ref_coords_arr(request):
    return request.param


_broadcastable_pairs = [
    (tuple(), tuple()),
    (tuple(), (2,)),
    ((5,), (5,)),
    ((2, 6), (2, 6)),
    ((2, 4, 3), (2, 4, 3)),
    ((3,), (4, 3)),
    ((4, 3), (3,)),
    ((1, 3), (4, 3)),
    ((4, 3), (1, 3)),
    ((4, 1, 2), (2, 1)),
    ((2, 1), (4, 1, 2)),
]


@pytest.fixture(params=_broadcastable_pairs)
def broadcastable_shapes(request):
    return request.param[0], request.param[1], np.broadcast_shapes(*request.param)


@pytest.fixture
def broadcastable_shape_triples(broadcastable_shapes):
    a, b, target = broadcastable_shapes

    num_entries = 3  # for triples

    # what each entry can be (this fixture fails if this is >= 10)
    candidates = [tuple(), a, b]
    # the tuple needs at least these
    requires = [1, 2]

    def is_valid(inds):
        # do checks:
        if any(r not in inds for r in requires):
            # we do not have a required entry
            return False

        # placeholder for more checks

        return True

    ind_gen = (
        [int(c) for c in np.base_repr(i, len(candidates))]
        for i in range(len(candidates) ** num_entries)
    )
    return target, ([candidates[i] for i in inds] for inds in ind_gen if is_valid(inds))


# ===================


def test_lagrange_poly_coefs1D(element):
    elem = element
    elem._lagrange_derivs = dict()
    elem._lagrange_derivs[0] = elem._lagrange_polys

    sigfigs = 5
    # we will use central finite difference which has O(h^2) error
    h = 10 ** -((sigfigs + 3) // 2)

    for deriv_order in range(elem.degree + 1):
        # verify that the lagrange derivatives are being set in the dictionary.
        P = elem.lagrange_poly1D(deriv_order)
        np.testing.assert_almost_equal(
            elem._lagrange_derivs[deriv_order],
            P,
            err_msg=(
                f"derivative L^{({deriv_order})} is not being set properly in the"
                " dictionary!"
            ),
        )
        assert set(elem._lagrange_derivs) == set(
            np.arange(deriv_order + 1)
        ), "Expected dictionary keys not found! Are they being improperly stored?"

    for deriv_order in range(1, elem.degree + 1):
        num_terms = elem.degree + 1 - deriv_order
        # polynomials of degree p are uniquely defined by their values at p+1 unique points. This is how we check equality.
        test_x = np.linspace(-1, 1, num_terms)

        # L_i^(deriv_order), compare to L_i^(...-1) with finite difference
        # polys are stored as P[i,k] : component c in term cx^k of poly P_i
        # L_i^(deriv_order-1)
        def L(x):
            return np.einsum(
                "ia,...a",
                elem.lagrange_poly1D(deriv_order - 1),
                np.expand_dims(x, -1) ** np.arange(num_terms + 1),
            )

        # L_i^(deriv_order)
        def Lp(x):
            return np.einsum(
                "ia,...a",
                elem.lagrange_poly1D(deriv_order),
                np.expand_dims(x, -1) ** np.arange(num_terms),
            )

        np.testing.assert_almost_equal(
            Lp(test_x),
            (L(test_x + h) - L(test_x - h)) / (2 * h),
            decimal=sigfigs,
            err_msg=(
                f"derivative L^({deriv_order}) does not match the central difference on"
                f" L^({deriv_order-1})!"
            ),
        )


def test_lagrange_evals1D(element, broadcastable_shapes):
    elem = element

    for deriv_order in range(elem.degree + 1):
        # verify that the lagrange derivatives are evaluated correctly
        P = elem.lagrange_poly1D(deriv_order)

        test_points = np.linspace(
            -1, 1, np.prod(broadcastable_shapes[1], dtype=int)
        ).reshape(broadcastable_shapes[1])
        test_indices = (
            np.arange(np.prod(broadcastable_shapes[0])).reshape(broadcastable_shapes[0])
            % P.shape[0]
        )
        if broadcastable_shapes[0] == tuple():
            test_indices = 0

        eval_pts = elem.lagrange_eval1D(deriv_order, test_indices, test_points)
        eval_alls = elem.lagrange_eval1D(deriv_order, test_indices, None)
        assert (
            eval_pts.shape == broadcastable_shapes[2]
        ), "Did not broadcast into the right shape!"
        assert eval_alls.shape == broadcastable_shapes[0] + (
            elem.num_nodes,
        ), "Did not broadcast into the right shape!"

        degp1 = elem.degree + 1 - deriv_order  # degree+1 of P
        Pknots = np.einsum(
            "ik,jk->ij", P, elem.knots[:, np.newaxis] ** np.arange(degp1)
        )
        np.testing.assert_almost_equal(eval_alls, Pknots[test_indices, :])

        test_points = np.broadcast_to(test_points, broadcastable_shapes[2])
        test_indices = np.broadcast_to(test_indices, broadcastable_shapes[2])

        it = np.nditer(eval_pts, flags=["multi_index"])
        for Px in it:
            ind = test_indices[it.multi_index]
            x = test_points[it.multi_index]
            np.testing.assert_almost_equal(
                Px,
                np.dot(P[ind, :], x ** np.arange(degp1)),
                err_msg=(
                    f"index {it.multi_index}: L_{ind}^({deriv_order}) ({x})"
                    " disagreement"
                ),
            )


# @pytest.mark.skip
def test_interpolate_field(element, ref_coords):
    field = element.basis_fields()
    x, y = ref_coords
    interp_vals = element.interpolate_field(field, x, y).coefficients
    # relies on correctness of lagrange_eval1D
    Lx = element.lagrange_eval1D(0, None, x=x)
    Ly = element.lagrange_eval1D(0, None, x=y)

    np.testing.assert_almost_equal(
        interp_vals,
        Lx[:, np.newaxis] * Ly[np.newaxis, :],
    )


def test_real_to_reference_interior(transformed_element, ref_coords):
    elem = transformed_element[0]
    points = transformed_element[1]
    transformation = transformed_element[2]

    true_pos = transformation(ref_coords)

    recover_ref = elem.locate_point(
        points.coefficients,
        true_pos[0],
        true_pos[1],
        tol=1e-10,
        ignore_out_of_bounds=True,
    )
    assert recover_ref[1], (
        "Test with ignore_out_of_bounds flag. Should be True for point being found"
        " (loss(recover_ref) < tol)."
    )
    np.testing.assert_almost_equal(
        recover_ref[0], ref_coords, err_msg="Test with ignore_out_of_bounds flag"
    )

    recover_ref = elem.locate_point(
        points.coefficients,
        true_pos[0],
        true_pos[1],
        tol=1e-10,
        ignore_out_of_bounds=False,
    )
    assert recover_ref[1], (
        "Test without ignore_out_of_bounds flag. Should be True for point being found"
        " (loss(recover_ref) < tol)."
    )
    np.testing.assert_almost_equal(
        recover_ref[0], ref_coords, err_msg="Test without ignore_out_of_bounds flag"
    )


def test_real_to_reference_degen_elem(element):
    points = element.reference_element_position_field().coefficients.copy()
    points[..., 0] = np.abs(points[..., 0])
    try:
        element.locate_point(points, 1e-10, 0.3)
    except spectral_element.DeformationGradient2DBadnessException as e:
        assert e.x == pytest.approx(0, abs=1e-5), "error should be on local x=0"
        return
    assert False, "Correct Exception not thrown!"


def test_integrate_field(
    transformed_element: tuple[spectral_element.SpectralElement2D, Field, Callable],
):
    element, pointfield, transform = transformed_element
    field = element.basis_fields()
    jac_scale = Field(
        element.basis_shape(), tuple(), field.coefficients[..., np.newaxis, np.newaxis]
    )
    result = element.integrate_field(pointfield, field, jac_scale)

    kronecker = field.coefficients.copy()
    kronecker *= element.weights[:, np.newaxis, np.newaxis, np.newaxis]
    kronecker *= element.weights[np.newaxis, :, np.newaxis, np.newaxis]

    def_grad = element.compute_field_gradient(pointfield)
    jac = np.abs(np.linalg.det(def_grad.coefficients))
    kronecker *= jac[:, :, np.newaxis, np.newaxis]

    np.testing.assert_almost_equal(result.coefficients, kronecker)


def test_integrate_basis_times_field(
    transformed_element: tuple[spectral_element.SpectralElement2D, Field, Callable],
):
    element, pointfield, transform = transformed_element
    field = element.basis_fields()
    jac_scale = Field(
        element.basis_shape(),
        tuple(),
        field.coefficients[:, :, np.newaxis, np.newaxis, :, :],
    )

    result = element.integrate_basis_times_field(
        pointfield, field, jacobian_scale=jac_scale
    )

    kronecker = field.coefficients.copy()

    triple_kronecker = np.einsum("abcd,cdef->abcdef", kronecker, kronecker)
    triple_kronecker *= element.weights[
        :, np.newaxis, np.newaxis, np.newaxis, np.newaxis, np.newaxis
    ]
    triple_kronecker *= element.weights[
        np.newaxis, :, np.newaxis, np.newaxis, np.newaxis, np.newaxis
    ]
    def_grad = element.compute_field_gradient(pointfield)
    jac = np.abs(np.linalg.det(def_grad.coefficients))
    triple_kronecker *= jac[:, :, np.newaxis, np.newaxis, np.newaxis, np.newaxis]

    np.testing.assert_almost_equal(result.coefficients, triple_kronecker)


def test_integrate_grad_basis_dot_field(
    transformed_element: tuple[spectral_element.SpectralElement2D, Field, Callable],
):
    element, pointfield, transform = transformed_element
    field = element.basis_fields()
    jac_scale = Field(
        element.basis_shape(),
        tuple(),
        field.coefficients[:, :, np.newaxis, np.newaxis, :, :],
    )
    e_field = Field(
        element.basis_shape(),
        (2,),
        np.eye(2)[:, *((np.newaxis,) * 4), np.newaxis, np.newaxis, :]
        * field.coefficients[np.newaxis, *((np.newaxis,) * 2), :, :, :, :, np.newaxis],
    )

    grad_basis = element.compute_field_gradient(
        element.basis_fields(), pointfield
    )  # ^g
    def_grad = element.compute_field_gradient(pointfield)  # [F]^g_l
    kronecker = field.coefficients.copy()

    # int(grad_basis * field * jac_scale)
    expect = np.einsum(
        "cdabg,abef,abhi,a,b,ab->gefhicd",
        grad_basis.coefficients,
        kronecker,
        kronecker,
        element.weights,
        element.weights,
        np.abs(np.linalg.det(def_grad.coefficients)),
    )

    result = element.integrate_grad_basis_dot_field(
        pointfield, e_field, jacobian_scale=jac_scale
    )

    np.testing.assert_almost_equal(result.coefficients, expect)


def meshquad(X, Y, F):
    """
    In case this is ever needed, this is a function to compute the integral on a 2D mesh.
    Integrals are estimated by estimating integrals on triangles, which are defined by
    indices ([i,j] - [i+1,j] - [i,j+1]) ([i+1,j+1] - [i+1,j] - [i,j+1])
    """

    def cross_mag(XA, YA, XB, YB):
        return np.abs(XA * YB - XB * YA)

    F00 = F[:-1, :-1]
    F01 = F[:-1, 1:]
    F10 = F[1:, :-1]
    F11 = F[1:, 1:]
    X00 = X[:-1, :-1]
    X01 = X[:-1, 1:]
    X10 = X[1:, :-1]
    X11 = X[1:, 1:]
    Y00 = Y[:-1, :-1]
    Y01 = Y[:-1, 1:]
    Y10 = Y[1:, :-1]
    Y11 = Y[1:, 1:]
    axes = np.arange(X.ndim)
    return np.sum(
        cross_mag(X10 - X00, Y10 - Y00, X01 - X00, Y01 - Y00) * (F00 + F01 + F10) / 6,
        axes,
    ) + np.sum(
        cross_mag(X10 - X11, Y10 - Y11, X01 - X11, Y01 - Y11) * (F11 + F01 + F10) / 6,
        axes,
    )


@pytest.mark.skip("we need to migrate this code over and generalize.")
def test_stiffness_matrix(transformed_element):
    elem, points, transformation = transformed_element
    # weights [i,j] times jacobian
    w = (
        elem.weights[:, np.newaxis]
        * elem.weights[np.newaxis, :]
        * np.abs(
            np.linalg.det(
                elem.interpolate_deformation_gradient(
                    points,
                    np.arange(elem.num_nodes),
                    np.arange(elem.num_nodes)[np.newaxis, :],
                )
            )
        )
    )

    # F(x_i,x_j) = delta_{im}delta_{jn}

    # equiv: # L_m(x_i)L_n(x_j)
    field = np.zeros((elem.num_nodes, elem.num_nodes, elem.num_nodes, elem.num_nodes))
    fieldshape = (elem.num_nodes, elem.num_nodes)
    enumeration = (
        np.arange(elem.num_nodes**2) % elem.num_nodes,
        np.arange(elem.num_nodes**2) // elem.num_nodes,
    )
    field[enumeration[0], enumeration[1], enumeration[0], enumeration[1]] = 1

    # [i,j, m,n, dim] partial_dim phi_{mn}(xi,xj)
    field_grad = elem.interpolate_field_gradient(
        field,
        elem.knots[:, np.newaxis],
        elem.knots[np.newaxis, :],
        pos_matrix=points,
        fieldshape=fieldshape,
    )

    # sum_{ij} w_{ij}( partial_dim phi_{mn}(xi,xj) )( partial_dim F_{ab}(xi,xj) )
    stiff = np.einsum("ij,ijmnd,ijabd->mnab", w, field_grad, field_grad)

    np.testing.assert_almost_equal(
        elem.integrate_grad_basis_dot_grad_field(points, field, fieldshape=fieldshape),
        stiff,
    )

    np.testing.assert_almost_equal(
        elem.integrate_grad_basis_dot_grad_field(points), np.einsum("mnmn->mn", stiff)
    )


@pytest.mark.parametrize("deg", [1, 2, 5, 10])
def test_gll_build(deg):
    x, w, L = spectral_element._build_GLL(deg)

    # verify shapes
    assert x.shape == (deg + 1,)
    assert w.shape == (deg + 1,)
    assert L.shape == (deg + 1, deg + 1)

    # verify quad degree
    for i in range(deg * 2 - 1):

        def F(x):
            return ((x + 1) / 2) ** i

        quad = sum(F(x) * w)
        res_true = 2 / (i + 1)
        assert quad == pytest.approx(res_true, abs=1e-8), (
            "GLL quadrature must be exact for polynomials up to degree 2n-1"
            f" ({2*deg - 1})! " + f"Failed at degree {i}."
        )

    # verify L
    evals = np.einsum(
        "ik,jk->ij", L, x[:, np.newaxis] ** np.arange(deg + 1)[np.newaxis, :]
    )
    np.testing.assert_almost_equal(evals, np.eye(deg + 1), decimal=8)


if __name__ == "__main__":
    pass



import collections.abc as colltypes
import typing

import numpy as np
from numpy.typing import NDArray

import fastfem.fields.numpy_similes as fnp
from fastfem.elements.element2d import StaticElement2D
from fastfem.elements.linear_simplex2d import LinearSimplex2D
from fastfem.fields.field import Field, FieldAxisIndex, ShapeComponent
from fastfem.fields.field_operations import assemble_field_add
from fastfem.mesh import Mesh, create_a_rectangle_mesh

atom = LinearSimplex2D()


class LinearSimplexMesh2D(StaticElement2D):
    def __init__(self, mesh: Mesh):
        nodes = {}
        elems = []
        # gather all nodes and elements in the above structures
        for component in mesh:
            if component.dimension == 2:
                for submesh in component.mesh:
                    if submesh.number_of_nodes_per_element != 3:
                        message = (
                            "LinearSimplexMesh2D operates on triangles!"
                            f" '{component.name}' has"
                            f" {submesh.number_of_nodes_per_element} nodes per element."
                        )
                        raise ValueError(message)
                    nodes.update(submesh.nodes)
                    elems.extend(submesh.elements.values())
        self.mesh = mesh
        self.num_nodes = len(nodes)
        self.num_elements = len(elems)
        node_index_to_basis_index = {}
        self.node_coords = np.empty((self.num_nodes, 2), dtype=float)
        self.element_node_indices = np.empty((self.num_elements, 3), dtype=int)
        # populate node coords, and obtain the node-index -> basis-index mapping
        for i, v in enumerate(nodes.items()):
            if abs(v[1][2]) > 1e-15:
                message = (
                    f"Node {i} must have a z-component of zero. Found"
                    f" {v[1][2]} instead."
                )
                raise ValueError(message)
            node_index_to_basis_index[v[0]] = i
            self.node_coords[i, :] = v[1][:2]

        # element [i,:] should be the basis indices of element i.
        for i in range(self.num_elements):
            for j in range(3):
                self.element_node_indices[i, j] = node_index_to_basis_index[elems[i][j]]

        self.position_field = Field((self.num_nodes,), (2,), self.node_coords)
        self.position_field_atomstack = self.to_atomstack(self.position_field)

    def to_atomstack(self, field: Field) -> Field:
        """Takes a field from the Mesh basis into the atom (LinearSimplex2D) field,
        where the field on each element `i` of the mesh corresponds to
        `atomstack.stack[i,...]`. That is, the basis shape becomes (3,), while an
        axis of size `num_elements` is prepended to the stack shape.

        Args:
            field (Field): The field to convert.

        Returns:
            Field: The field as represented by a stack of LinearSimplex2D fields.
        """
        self._verify_field_compatibilities(field)
        field = field.broadcast_to_shape(
            field.stack_shape, self.basis_shape(), field.point_shape
        )
        return fnp.moveaxis(
            field.basis[self.element_node_indices],
            (ShapeComponent.BASIS, 0),
            (ShapeComponent.STACK, 0),
        )

    def from_atomstack_accumulate(self, field: Field) -> Field:
        """Takes a field stack from the atomic (LinearSimplex2D) bases into the Mesh basis,
        where the field on each element `i` of the mesh corresponds to
        `atomstack.stack[i,...]`. The leading stack component is contracted. Repeated
        nodes are added, so this method is used to add integrals of shape functions.

        Args:
            field (Field): The field to convert.

        Returns:
            Field: The field in the mesh basis.
        """
        return assemble_field_add(field, self.element_node_indices, self.basis_shape())

    @typing.override
    def basis_shape(self) -> tuple[int, ...]:
        return (self.num_nodes,)

    @typing.override
    def _interpolate_field(self, field: Field, X: NDArray, Y: NDArray) -> Field:
        raise NotImplementedError

    @typing.override
    def _integrate_field(self, field: Field, jacobian_scale: Field) -> Field:
        return fnp.sum(
            atom._integrate_field(
                self.position_field_atomstack.stack[
                    :, *(np.newaxis for _ in field.stack_shape)
                ],
                self.to_atomstack(field),
                self.to_atomstack(jacobian_scale),
            ),
            FieldAxisIndex(ShapeComponent.STACK, 0),
        )

    @typing.override
    def _integrate_basis_times_field(
        self, field: Field, indices: colltypes.Sequence | None, jacobian_scale: Field
    ) -> Field:
        if indices is not None:
            message = "Index assembly not yet implemented"
            raise NotImplementedError(message)
        return self.from_atomstack_accumulate(
            atom._integrate_basis_times_field(
                self.position_field_atomstack.stack[
                    :, *(np.newaxis for _ in field.stack_shape)
                ],
                self.to_atomstack(field),
                indices=None,
                jacobian_scale=self.to_atomstack(jacobian_scale),
            )
        )

    @typing.override
    def _integrate_grad_basis_dot_field(
        self, field: Field, indices: colltypes.Sequence | None, jacobian_scale: Field
    ) -> Field:
        if indices is not None:
            message = "Index assembly not yet implemented"
            raise NotImplementedError(message)
        return self.from_atomstack_accumulate(
            atom._integrate_grad_basis_dot_field(
                self.position_field_atomstack.stack[
                    :, *(np.newaxis for _ in field.stack_shape)
                ],
                self.to_atomstack(field),
                indices=None,
                jacobian_scale=self.to_atomstack(jacobian_scale),
            ),
        )

    @typing.override
    def _integrate_grad_basis_dot_grad_field(
        self, field: Field, indices: colltypes.Sequence | None, jacobian_scale: Field
    ) -> Field:
        if indices is not None:
            message = "Index assembly not yet implemented"
            raise NotImplementedError(message)
        return self.from_atomstack_accumulate(
            atom._integrate_grad_basis_dot_grad_field(
                self.position_field_atomstack.stack[
                    :, *(np.newaxis for _ in field.stack_shape)
                ],
                self.to_atomstack(field),
                indices=None,
                jacobian_scale=self.to_atomstack(jacobian_scale),
            ),
        )


if __name__ == "__main__":
    elem = LinearSimplexMesh2D(
        create_a_rectangle_mesh(
            horizontal_length=1,
            vertical_length=1,
            nodes_in_horizontal_direction=10,
            nodes_in_vertical_direction=10,
            element_type="triangle",
            file_name=None,
        )
    )



from .linear_simplex2d import LinearSimplex2D
from .linear_simplex_mesh2d import LinearSimplexMesh2D
from .spectral_element2d import SpectralElement2D

__all__ = [
    "LinearSimplex2D",
    "LinearSimplexMesh2D",
    "SpectralElement2D",
]



from collections.abc import Callable, Iterable
from typing import Literal

import numpy as np
import pytest
from numpy.typing import NDArray

from fastfem.fields.field import Field


def transform_posmatrix_linearly(pos_matrix, mod, *args):
    if mod == "translate":
        if len(args) < 2:
            raise ValueError(f"modifier '{mod}' expects 2 arguments! (dx,dy)")
        vec = np.array([args[0], args[1]])
        pos_matrix = pos_matrix + vec  # no in-place op, since we want a copy
    elif mod == "rotate":
        if len(args) < 1:
            raise ValueError(f"modifier '{mod}' expects 1 argument! (angle)")
        t = args[0]
        rotmat = np.array([[np.cos(t), -np.sin(t)], [np.sin(t), np.cos(t)]])
        pos_matrix = (rotmat @ np.expand_dims(pos_matrix, -1)).squeeze(-1)
    elif mod == "scale":
        if len(args) < 2:
            raise ValueError(f"modifier '{mod}' expects 2 arguments! (scalex,scaley)")
        vec = np.array([args[0], args[1]])
        pos_matrix = pos_matrix * vec  # no in-place op, since we want a copy
    elif mod == "lin_trans":
        if len(args) < 4:
            raise ValueError(f"modifier '{mod}' expects 4 arguments! (m00,m01,m10,m11)")
        A = np.array([[args[0], args[1]], [args[2], args[3]]])
        pos_matrix = (A @ np.expand_dims(pos_matrix, -1)).squeeze(-1)
    else:
        raise ValueError(f"'{mod}' not acceptable element modifier!")
    return pos_matrix


_PRESET_TRANSFORMS_AFFINE = {
    "id": lambda x: x,
    "translated": lambda x: transform_posmatrix_linearly(x, "translate", 5, -2),
    "rotated": lambda x: transform_posmatrix_linearly(x, "rotate", 1),
    "x-scaled": lambda x: transform_posmatrix_linearly(x, "scale", 2, 1),
    "y-scaled": lambda x: transform_posmatrix_linearly(x, "scale", 1, 2),
    "combo1": lambda x: transform_posmatrix_linearly(
        transform_posmatrix_linearly(x, "lin_trans", 2, 1, -1, 1), "translate", -4, 2
    ),
    "combo2": lambda x: transform_posmatrix_linearly(
        transform_posmatrix_linearly(x, "lin_trans", 0.5, 1.3, 10, 0.3),
        "translate",
        300,
        600,
    ),
}


class AffineTransformHandler:
    def set_target_type(
        self,
        T: Callable[[NDArray], NDArray],
        target_type: Literal["point"] | Literal["vector"] | Literal["form"],
    ) -> Callable[[NDArray], NDArray]:
        if target_type == "point":
            return T
        if target_type == "vector":
            return lambda x: T(x) - T(np.zeros(2))
        if target_type == "form":
            shift = T(np.zeros(2))
            A_Tinv = np.linalg.inv(T(np.eye(2)) - shift).T
            return lambda x: (A_Tinv @ np.expand_dims(x, -1)).squeeze(-1)
        raise ValueError(
            f"target_type {target_type} not known. Must be 'point', 'vector', or 'form'"
        )

    def iterate_presets(
        self,
        target_type: Literal["point"] | Literal["vector"] | Literal["form"] = "point",
    ) -> Iterable[tuple[str, Callable[[NDArray], NDArray]]]:
        for k, T in _PRESET_TRANSFORMS_AFFINE.items():
            yield k, self.set_target_type(T, target_type)

    def __iter__(self):
        yield from _PRESET_TRANSFORMS_AFFINE.items()


@pytest.fixture
def affine_transforms():
    return AffineTransformHandler()


@pytest.fixture(scope="module", params=_PRESET_TRANSFORMS_AFFINE.keys())
def transformation(request):
    name = request.param
    return _PRESET_TRANSFORMS_AFFINE[name]


@pytest.fixture(scope="module", params=[0, 1, 2, 3])
def transform_stack(request):
    transforms = _PRESET_TRANSFORMS_AFFINE.values()
    # param is number of dims for element position array
    ndims = request.param
    stackshape = tuple(3 for _ in range(ndims))
    stacksize = np.prod(stackshape, dtype=int)

    def stack_transform(x: Field):
        y = np.empty(x.basis_shape + stackshape + x.stack_shape + x.point_shape)
        transformed = [f(x.coefficients) for f in transforms]
        for i in range(stacksize):
            y[..., *np.unravel_index(i, stackshape), ...] = transformed[
                i % len(transforms)
            ]
        return Field(x.basis_shape, x.point_shape, y)

    return stack_transform



import dataclasses
import itertools
import typing
from dataclasses import dataclass
from enum import IntEnum
from types import ModuleType
from typing import Literal, Union

import jax
import numpy as _np  # noqa: ICN001
from numpy.typing import ArrayLike, NDArray

jnp = _np  # this line is just to make type hinting not complain about np()
HAS_JAX = True
try:
    import jax.numpy as jnp
except ImportError:
    HAS_JAX = False


def np_or_jnp(*use_jax) -> ModuleType:
    if any((entry.use_jax if isinstance(entry, Field) else entry) for entry in use_jax):
        if HAS_JAX:
            return jnp
        message = "JAX was not found, but was requested!"
        raise ValueError(message)
    return _np


np = _np


def _is_broadcastable(base: tuple[int, ...], *shapes: tuple[int, ...]) -> bool:
    """Checks if shapes are broadcastable into base by numpy broadcasting rules.

    Args:
        base (tuple[int,...]): the first (target) shape
        *shapes (tuple[int,...]): the shapes to be broadcasted

    Returns:
        bool: `True` if the shapes are compatible, and false otherwise.
    """
    return all(
        map(
            lambda x: x[0] is not None
            and all(xi == x[0] or xi == 1 or xi is None for xi in x[1:]),
            itertools.zip_longest(
                reversed(base), *(reversed(shape) for shape in shapes), fillvalue=None
            ),
        )
    )


def _is_compatible(*shapes: tuple[int, ...]) -> bool:
    """Checks if shapes are compatible by numpy broadcasting rules.

    Returns:
        bool: `True` if the shapes are compatible, and false otherwise.
    """
    return all(
        map(
            lambda xi: all(  # every shape[i] in x[i] must be compatible
                map(  # recover compatibility boolean
                    lambda fjxi: fjxi[1],
                    # fj(x[i]) = (axsize , axsize compatible with x[i][j]);  j = 0,...
                    itertools.accumulate(
                        xi,
                        func=lambda a, b: (
                            b if b != 1 else a[0],
                            (a[0] == b or a[0] == 1 or b == 1),
                        ),
                        initial=(1, True),
                    ),
                )
            ),
            # x[i] = (shape[i] for shape in shapes) : i = 0,...
            itertools.zip_longest(*(reversed(shape) for shape in shapes), fillvalue=1),
        )
    )


class FieldShapeError(Exception):
    pass


class ShapeComponent(IntEnum):
    STACK = 0
    BASIS = 1
    POINT = 2


@dataclass(frozen=True)
class FieldAxisIndex:
    component: ShapeComponent
    index: int

    @typing.overload
    def __getitem__(self, ind: Literal[0]) -> ShapeComponent: ...
    @typing.overload
    def __getitem__(self, ind: Literal[1]) -> int: ...

    def __getitem__(self, ind):
        if ind == 0:
            return self.component
        return self.index


FieldAxisIndexType = tuple[ShapeComponent, int] | FieldAxisIndex


def _verify_is_permutation(p: tuple[int, ...]) -> None:
    if not isinstance(p, tuple):
        message = "shape_order is not a permutation! (must be a tuple)"
        raise FieldShapeError(message)
    n = len(p)  # size of the permutation
    exists = [False] * n
    for i in p:
        if not isinstance(i, int):
            message = "shape_order is not a permutation! (all entries must be integers)"
            raise FieldShapeError(message)
        exists[i] = True

    if not all(exists):
        message = "shape_order is not a permutation! (must be a bijection)"
        raise FieldShapeError(message)


def _invert_permutation(p: tuple[int, ...]) -> tuple[int, ...]:
    # assume p is a permutation
    pinv = [-1] * len(p)
    for i, k in enumerate(p):
        pinv[k] = i
    return tuple(pinv)


def _reshape(
    field: "Field",
    component_selector: ShapeComponent,
    shape: int | tuple[int],
    order: Literal["C", "F", "A"] = "C",
    copy: bool | None = None,
) -> "Field":
    """This attempts to replicate the numpy "reshape" function.
    https://numpy.org/doc/stable/reference/generated/numpy.reshape.html

    reshape() applies numpy "reshape" to a given component. This function is also called
    when using the field accessor reshape methods. That is, `reshape(field,BASIS,s)` is
    the same as `field.basis.reshape(s)`
    Args:
        field (Field): The field to reshape
        component_selector (ShapeComponent): Which component to reshape.
        shape (int | tuple[int]): The target shape of the component. For any integer i,
            i is equivalent to (i,)
        order ({'C','F','A'}, optional): See the numpy documentation. Defaults to 'C'.
        copy (bool | None, optional): See the numpy documentation. Defaults to None.

    Raises:
        ValueError: when a copy operation is required, but `copy` is False.
    Returns:
        Field: The reshaped field. This is always a new object, but if data is copied,
        the underlying array is a view of the original.
    """
    # if not isinstance(shape, tuple):
    #     shape = (shape,)
    # start_ind = field._axis_field_to_numpy(FieldAxisIndex(component_selector, 0))
    # end_ind = field._axis_field_to_numpy(FieldAxisIndex(component_selector, -1)) + 1
    # shapes = [()] * 3
    # shapes[ShapeComponent.BASIS] = field.basis_shape
    # shapes[ShapeComponent.POINT] = field.point_shape
    # shapes[ShapeComponent.STACK] = field.stack_shape
    # shapes[component_selector] = shape
    # np_target_shape = shapes[0] + shapes[1] + shapes[2]
    # coefs = np_or_jnp(field.use_jax).reshape(
    #         field.coefficients, np_target_shape, order=order, copy=copy
    #     )
    # shape_ = coefs.shape[start_ind:end_ind]
    # return Field(
    #     shape_ if component_selector == ShapeComponent.BASIS else field.basis_shape,
    #     shape_ if component_selector == ShapeComponent.POINT else field.point_shape,
    #     coefs,
    #     shape_order=field.shape_order,
    #     use_jax=field.use_jax,
    # )
    if not isinstance(shape, tuple):
        shape = (shape,)
    # beginning of reshaped section
    start_ind = field._component_offset(component_selector)
    # end of reshaped section (excl)
    end_ind = start_ind + len(shape)
    shapes = [()] * 3
    shapes[field.shape_order[ShapeComponent.BASIS]] = field.basis_shape  # type: ignore
    shapes[field.shape_order[ShapeComponent.POINT]] = field.point_shape  # type: ignore
    shapes[field.shape_order[ShapeComponent.STACK]] = field.stack_shape  # type: ignore
    shapes[field.shape_order[component_selector]] = shape  # type: ignore
    np_target_shape = shapes[0] + shapes[1] + shapes[2]
    coefs = np_or_jnp(field.use_jax).reshape(
        field.coefficients, np_target_shape, order=order, copy=copy
    )
    shape_ = coefs.shape[start_ind:end_ind]
    return Field(
        shape_ if component_selector == ShapeComponent.BASIS else field.basis_shape,
        shape_ if component_selector == ShapeComponent.POINT else field.point_shape,
        coefs,
        shape_order=field.shape_order,
        use_jax=field.use_jax,
    )


class FieldConstructionError(FieldShapeError):
    """Called when constructing a field fails."""

    def __init__(self, basis_shape, point_shape, coeff_shape, shape_order, hint=None):
        errmsg = (
            f"Cannot construct Field object with basis_shape {basis_shape}, point_shape"
            f" {point_shape} given the coefficient shape {coeff_shape} and shape order"
            f" {shape_order}."
        )
        if hint is not None:
            errmsg += f" ({hint})"
        super().__init__(errmsg)


class FieldBasisAccessor:
    """This class is the type returned in the Field __get_attr__ for basis access.
    The sole purpose of this class is to provide the syntax for basis-targeted slicing
    in the same format as numpy.
    """

    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, slices):
        # do validation
        basis = _np.broadcast_to(0, self.parent.basis_shape)
        new_basis_shape = basis[slices].shape
        slicepad = (slice(None),) * (
            (
                len(self.parent.stack_shape)
                if self.parent.shape_order[ShapeComponent.STACK]
                < self.parent.shape_order[ShapeComponent.BASIS]
                else 0
            )
            + (
                len(self.parent.point_shape)
                if self.parent.shape_order[ShapeComponent.POINT]
                < self.parent.shape_order[ShapeComponent.BASIS]
                else 0
            )
        )
        if not isinstance(slices, tuple):
            slices = (slices,)
        return Field(
            new_basis_shape,
            self.parent.point_shape,
            self.parent.coefficients[*slicepad, *slices],
        )

    def reshape(
        self,
        shape: int | tuple[int],
        order: Literal["C", "F", "A"] = "C",
        copy: bool | None = None,
    ) -> "Field":
        """This attempts to replicate the numpy "reshape" function.
        https://numpy.org/doc/stable/reference/generated/numpy.reshape.html

        Delegates to numpy_similes.reshape()

        Args:
            field (Field): The field to reshape
            component_selector (ShapeComponent): Which component to reshape.
            shape (int | tuple[int]): The target shape of the component. For any integer i,
                i is equivalent to (i,)
            order ({'C','F','A'}, optional): See the numpy documentation. Defaults to 'C'.
            copy (bool | None, optional): See the numpy documentation. Defaults to None.

        Raises:
            ValueError: when a copy operation is required, but `copy` is False.
        Returns:
            Field: The reshaped field. This is always a new object, but if data is copied,
            the underlying array is a view of the original.
        """
        return _reshape(self.parent, ShapeComponent.BASIS, shape, order, copy)


class FieldStackAccessor:
    """This class is the type returned in the Field __get_attr__ for stack access.
    The sole purpose of this class is to provide the syntax for stack-targeted slicing
    in the same format as numpy.
    """

    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, slices):
        # do validation
        stack = np.broadcast_to(0, self.parent.stack_shape)
        new_stack_shape = stack[slices].shape  # noqa: F841
        slicepad = (slice(None),) * (
            (
                len(self.parent.basis_shape)
                if self.parent.shape_order[ShapeComponent.BASIS]
                < self.parent.shape_order[ShapeComponent.STACK]
                else 0
            )
            + (
                len(self.parent.point_shape)
                if self.parent.shape_order[ShapeComponent.POINT]
                < self.parent.shape_order[ShapeComponent.STACK]
                else 0
            )
        )
        if not isinstance(slices, tuple):
            slices = (slices,)
        return Field(
            self.parent.basis_shape,
            self.parent.point_shape,
            self.parent.coefficients[*slicepad, *slices],
        )

    def reshape(
        self,
        shape: int | tuple[int],
        order: Literal["C", "F", "A"] = "C",
        copy: bool | None = None,
    ) -> "Field":
        """This attempts to replicate the numpy "reshape" function.
        https://numpy.org/doc/stable/reference/generated/numpy.reshape.html

        Delegates to numpy_similes.reshape()

        Args:
            field (Field): The field to reshape
            component_selector (ShapeComponent): Which component to reshape.
            shape (int | tuple[int]): The target shape of the component. For any integer i,
                i is equivalent to (i,)
            order ({'C','F','A'}, optional): See the numpy documentation. Defaults to 'C'.
            copy (bool | None, optional): See the numpy documentation. Defaults to None.

        Raises:
            ValueError: when a copy operation is required, but `copy` is False.
        Returns:
            Field: The reshaped field. This is always a new object, but if data is copied,
            the underlying array is a view of the original.
        """
        return _reshape(self.parent, ShapeComponent.STACK, shape, order, copy)


class FieldPointAccessor:
    """This class is the type returned in the Field __get_attr__ for pointwise field-element
    access.
    The sole purpose of this class is to provide the syntax for field-targeted slicing
    in the same format as numpy.
    """

    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, slices):
        # do validation
        element = np.broadcast_to(0, self.parent.point_shape)
        new_point_shape = element[slices].shape
        slicepad = (slice(None),) * (
            (
                len(self.parent.basis_shape)
                if self.parent.shape_order[ShapeComponent.BASIS]
                < self.parent.shape_order[ShapeComponent.POINT]
                else 0
            )
            + (
                len(self.parent.stack_shape)
                if self.parent.shape_order[ShapeComponent.STACK]
                < self.parent.shape_order[ShapeComponent.POINT]
                else 0
            )
        )
        if not isinstance(slices, tuple):
            slices = (slices,)
        return Field(
            self.parent.basis_shape,
            new_point_shape,
            self.parent.coefficients[*slicepad, *slices],
        )

    def reshape(
        self,
        shape: int | tuple[int],
        order: Literal["C", "F", "A"] = "C",
        copy: bool | None = None,
    ) -> "Field":
        """This attempts to replicate the numpy "reshape" function.
        https://numpy.org/doc/stable/reference/generated/numpy.reshape.html

        Delegates to numpy_similes.reshape()

        Args:
            field (Field): The field to reshape
            component_selector (ShapeComponent): Which component to reshape.
            shape (int | tuple[int]): The target shape of the component. For any integer i,
                i is equivalent to (i,)
            order ({'C','F','A'}, optional): See the numpy documentation. Defaults to 'C'.
            copy (bool | None, optional): See the numpy documentation. Defaults to None.

        Raises:
            ValueError: when a copy operation is required, but `copy` is False.
        Returns:
            Field: The reshaped field. This is always a new object, but if data is copied,
            the underlying array is a view of the original.
        """
        return _reshape(self.parent, ShapeComponent.POINT, shape, order, copy)


@dataclass(eq=False, frozen=True, unsafe_hash=False, init=False)
class Field:
    """
    A class responsible for storing fields on elements as an `NDArray` of coefficients.
    There are 3 relevant shapes / axis sets to a field:

    - `basis_shape` - The shape of the basis. These axes represent the multi-index for
            the basis function.

    - `stack_shape` - The shape of the element stack. These axes represent the
            multi-index for the element.

    - `point_shape` - The shape of the field. These axes represent the pointwise,
            per-element tensor index.

    The shape of `coefficients` will be some permutation of
    `stack_shape + point_shape + basis_shape`. The order is specified by `shape_order`,
    which is a 3-tuple `(stack_location, field_location, basis_location)`, where each
    entry is an integer specifying the position relative to the other two shapes.
    """

    basis_shape: tuple[int, ...]
    stack_shape: tuple[int, ...]
    point_shape: tuple[int, ...]
    coefficients: NDArray | jax.Array
    shape_order: tuple[int, int, int] = dataclasses.field(repr=False, init=False)
    shape_order_inverse: tuple[ShapeComponent, ShapeComponent, ShapeComponent] = (
        dataclasses.field(repr=False, init=False)
    )
    use_jax: bool = dataclasses.field(repr=False, init=False)

    def __init__(
        self,
        basis_shape: tuple[int, ...],
        point_shape: tuple[int, ...],
        coefficients: ArrayLike,
        shape_order: tuple[int, int, int] = (0, 1, 2),
        use_jax: bool | None = None,
    ):
        _verify_is_permutation(shape_order)
        if not isinstance(coefficients, np.ndarray) and not isinstance(
            coefficients, jax.Array
        ):
            coefficients = np.array(coefficients)
        if use_jax is None:
            use_jax = isinstance(coefficients, jax.Array)
        cshape_orig = np.shape(coefficients)
        if len(cshape_orig) < len(basis_shape) + len(point_shape):
            coefficients = coefficients[
                *(
                    (np.newaxis,)
                    * (len(basis_shape) + len(point_shape) - len(cshape_orig))
                ),
                ...,
            ]
            cshape = np.shape(coefficients)
        else:
            cshape = cshape_orig

        # here, coefficients is at least as large as basis and field shapes combined

        # we need to place two markers to index the separations between basis, field,
        # and stack shapes; start with basis_shape (if not in middle)
        stack_start = 0
        stack_end = len(cshape)

        def cshape_slice_positives(a, b):
            return cshape[a:b]

        def cshape_slice_negatives(a, b):
            return cshape[a : (b if b != 0 else None)] if a != 0 else ()

        if shape_order[ShapeComponent.BASIS] == 0:
            if not _is_broadcastable(
                basis_shape, cshape_slice_positives(0, len(basis_shape))
            ):
                raise FieldConstructionError(
                    basis_shape,
                    point_shape,
                    cshape_orig,
                    shape_order,
                    hint="basis_shape cannot be broadcasted at the beginning",
                )
            stack_start = len(basis_shape)
        elif shape_order[ShapeComponent.BASIS] == 2:
            if not _is_broadcastable(
                basis_shape, cshape_slice_negatives(-len(basis_shape), 0)
            ):
                raise FieldConstructionError(
                    basis_shape,
                    point_shape,
                    cshape_orig,
                    shape_order,
                    hint="basis_shape cannot be broadcasted at the end",
                )
            stack_end -= len(basis_shape)
        # then do point_shape
        if shape_order[ShapeComponent.POINT] == 0:
            if not _is_broadcastable(
                point_shape, cshape_slice_positives(0, len(point_shape))
            ):
                raise FieldConstructionError(
                    basis_shape,
                    point_shape,
                    cshape_orig,
                    shape_order,
                    hint="point_shape cannot be broadcasted at the beginning",
                )
            # if basis_shape was in center, we now have the right offset for it
            if shape_order[ShapeComponent.BASIS] == 1:
                if not _is_broadcastable(
                    basis_shape,
                    cshape_slice_positives(
                        len(point_shape), (len(basis_shape) + len(point_shape))
                    ),
                ):
                    raise FieldConstructionError(
                        basis_shape,
                        point_shape,
                        cshape_orig,
                        shape_order,
                        hint="basis_shape cannot be broadcasted in the center",
                    )
                stack_start = len(basis_shape) + len(point_shape)
            else:
                stack_start = len(point_shape)
        elif shape_order[ShapeComponent.POINT] == 2:
            if not _is_broadcastable(
                point_shape, cshape_slice_negatives(-len(point_shape), 0)
            ):
                raise FieldConstructionError(
                    basis_shape,
                    point_shape,
                    cshape_orig,
                    shape_order,
                    hint="point_shape cannot be broadcasted at the end",
                )
            # if basis_shape was in center, we now have the right offset for it
            if shape_order[ShapeComponent.BASIS] == 1:
                if not _is_broadcastable(
                    basis_shape,
                    cshape_slice_negatives(
                        -(len(basis_shape) + len(point_shape)), -len(point_shape)
                    ),
                ):
                    raise FieldConstructionError(
                        basis_shape,
                        point_shape,
                        cshape_orig,
                        shape_order,
                        hint="basis_shape cannot be broadcasted in the center",
                    )
                stack_end -= len(basis_shape) + len(point_shape)
            else:
                stack_end -= len(point_shape)
        elif shape_order[ShapeComponent.POINT] == 1:
            # cases by basis_location
            if shape_order[ShapeComponent.BASIS] == 0:
                if not _is_broadcastable(
                    point_shape,
                    cshape_slice_positives(
                        len(basis_shape), (len(basis_shape) + len(point_shape))
                    ),
                ):
                    raise FieldConstructionError(
                        basis_shape,
                        point_shape,
                        cshape_orig,
                        shape_order,
                        hint="point_shape cannot be broadcasted in the center",
                    )
                stack_start += len(point_shape)
            else:
                if not _is_broadcastable(
                    point_shape,
                    cshape_slice_negatives(
                        -(len(basis_shape) + len(point_shape)), -len(basis_shape)
                    ),
                ):
                    raise FieldConstructionError(
                        basis_shape,
                        point_shape,
                        cshape_orig,
                        shape_order,
                        hint="point_shape cannot be broadcasted in the center",
                    )
                stack_end -= len(point_shape)

        stack_shape = cshape[stack_start:stack_end]
        shapes: list[tuple[int, ...]] = [()] * 3
        shapes[shape_order[ShapeComponent.BASIS]] = basis_shape
        shapes[shape_order[ShapeComponent.STACK]] = stack_shape
        shapes[shape_order[ShapeComponent.POINT]] = point_shape
        object.__setattr__(
            self,
            "coefficients",
            np_or_jnp(use_jax).broadcast_to(
                coefficients, shapes[0] + shapes[1] + shapes[2]
            ),
        )
        object.__setattr__(self, "basis_shape", basis_shape)
        object.__setattr__(self, "point_shape", point_shape)
        object.__setattr__(self, "stack_shape", stack_shape)
        object.__setattr__(self, "shape_order", shape_order)
        object.__setattr__(
            self, "shape_order_inverse", _invert_permutation(shape_order)
        )
        object.__setattr__(self, "use_jax", use_jax)

    @typing.overload
    def __getattr__(
        self, name: Literal["shape"]
    ) -> tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]: ...
    @typing.overload
    def __getattr__(self, name: Literal["basis"]) -> FieldBasisAccessor: ...
    @typing.overload
    def __getattr__(self, name: Literal["stack"]) -> FieldStackAccessor: ...
    @typing.overload
    def __getattr__(self, name: Literal["point"]) -> FieldPointAccessor: ...

    def __getattr__(self, name):
        if name == "shape":
            return (self.stack_shape, self.basis_shape, self.point_shape)
        if name == "basis":
            return FieldBasisAccessor(self)
        if name == "stack":
            return FieldStackAccessor(self)
        if name == "point":
            return FieldPointAccessor(self)
        raise AttributeError

    def __add__(self, other: Union["Field", float]):
        if isinstance(other, Field):
            a, b = Field.broadcast_fields_full(self, other)
            return Field(
                a.basis_shape,
                a.point_shape,
                a.coefficients + b.coefficients,
                shape_order=a.shape_order,
                use_jax=a.use_jax or b.use_jax,
            )
        if isinstance(other, float):
            return Field(
                self.basis_shape,
                self.point_shape,
                self.coefficients + other,
                shape_order=self.shape_order,
                use_jax=self.use_jax,
            )
        raise NotImplementedError

    def __radd__(self, other: Union["Field", float]):
        return self.__add__(other)

    def __sub__(self, other: Union["Field", float]):
        if isinstance(other, Field):
            a, b = Field.broadcast_fields_full(self, other)
            return Field(
                a.basis_shape,
                a.point_shape,
                a.coefficients - b.coefficients,
                shape_order=a.shape_order,
                use_jax=a.use_jax or b.use_jax,
            )
        if isinstance(other, float):
            return Field(
                self.basis_shape,
                self.point_shape,
                self.coefficients - other,
                shape_order=self.shape_order,
                use_jax=self.use_jax,
            )
        raise NotImplementedError

    def __rsub__(self, other: Union["Field", float]):
        if isinstance(other, Field):
            a, b = Field.broadcast_fields_full(self, other)
            return Field(
                a.basis_shape,
                a.point_shape,
                b.coefficients - a.coefficients,
                shape_order=a.shape_order,
                use_jax=a.use_jax or b.use_jax,
            )
        if isinstance(other, float):
            return Field(
                self.basis_shape,
                self.point_shape,
                other - self.coefficients,
                shape_order=self.shape_order,
                use_jax=self.use_jax,
            )
        raise NotImplementedError

    def __mul__(self, other: Union["Field", float]):
        if isinstance(other, Field):
            a, b = Field.broadcast_fields_full(self, other)
            return Field(
                a.basis_shape,
                a.point_shape,
                a.coefficients * b.coefficients,
                shape_order=a.shape_order,
                use_jax=a.use_jax or b.use_jax,
            )
        if isinstance(other, float):
            return Field(
                self.basis_shape,
                self.point_shape,
                self.coefficients * other,
                shape_order=self.shape_order,
                use_jax=self.use_jax,
            )
        raise NotImplementedError

    def __rmul__(self, other: Union["Field", float]):
        return self.__mul__(other)

    def __neg__(self):
        return Field(
            self.basis_shape,
            self.point_shape,
            -self.coefficients,
            shape_order=self.shape_order,
            use_jax=self.use_jax,
        )

    def get_shape(self, component: ShapeComponent) -> tuple[int, ...]:
        """Recovers the shape of the specified component. This shape is in the same
        format as numpy.shape, that is a tuple.

        Args:
            component (ShapeComponent): The component to sample

        Returns:
            tuple[int,...]: The shape of the specified component.
        """
        return (
            self.basis_shape
            if component == ShapeComponent.BASIS
            else (
                self.stack_shape
                if component == ShapeComponent.STACK
                else self.point_shape
            )
        )

    def _component_offset(self, component: ShapeComponent) -> int:
        """Recovers the axis index of the start of a component.

        Args:
            component (ShapeComponent): The component to sample.

        Returns:
            int: the index of the first axis of the component
        """
        ind = 0
        prec = self.shape_order[component]
        if self.shape_order[ShapeComponent.BASIS] < prec:
            ind += len(self.basis_shape)
        if self.shape_order[ShapeComponent.STACK] < prec:
            ind += len(self.stack_shape)
        if self.shape_order[ShapeComponent.POINT] < prec:
            ind += len(self.point_shape)
        return ind

    def broadcast_to_shape(
        self,
        stack_shape: tuple[int, ...] | None,
        basis_shape: tuple[int, ...] | None,
        point_shape: tuple[int, ...] | None,
        complete_broadcast=True,
    ) -> "Field":
        """This function is related to the numpy broadcast_to function.
        https://numpy.org/doc/stable/reference/generated/numpy.broadcast_to.html#numpy.broadcast_to

        The shape of the desired array is given as separate tuples for each component.
        Instead of the `subok` optional argument, the returned field will always have
        the same coefficient array type. Additionally, the value of `use_jax` is inherited.

        Args:
            stack_shape (tuple[int, ...] | None): The shape for the stack shape to be
                broadcasted to, or None, if the shape should be kept as-is.
            basis_shape (tuple[int, ...] | None): The shape for the basis shape to be
                broadcasted to, or None, if the shape should be kept as-is.
            point_shape (tuple[int, ...] | None): The shape for the field shape to be
                broadcasted to, or None, if the shape should be kept as-is.
            complete_broadcast (bool, optional): If False, only dimensions of size 1
                are added. When true, the shape of the field precisely matches.

        Raises:
            FieldShapeError: if the broadcast cannot be done by standard numpy
                broadcasting rules in each component. Note that broadcasting the basis
                shape is permitted, beyond standard compatibility rules.

        Returns:
            Field: The broadcasted field.
        """
        if (
            (
                basis_shape is not None
                and not _is_broadcastable(basis_shape, self.basis_shape)
            )
            or (
                stack_shape is not None
                and not _is_broadcastable(stack_shape, self.stack_shape)
            )
            or (
                point_shape is not None
                and not _is_broadcastable(point_shape, self.point_shape)
            )
        ):
            message = (
                f"Cannot broadcast field of shape {self.shape} into"
                f" shape {(stack_shape,basis_shape,point_shape)}"
            )
            raise FieldShapeError(message)
        slices: list[typing.Any] = [None, None, None]
        shapes: list[typing.Any] = [None, None, None]
        slices[self.shape_order[ShapeComponent.BASIS]] = (
            itertools.chain(
                (np.newaxis for _ in range(len(basis_shape) - len(self.basis_shape))),
                (slice(None) for _ in range(len(self.basis_shape))),
            )
            if basis_shape is not None
            else (slice(None) for _ in range(len(self.basis_shape)))
        )
        slices[self.shape_order[ShapeComponent.STACK]] = (
            itertools.chain(
                (np.newaxis for _ in range(len(stack_shape) - len(self.stack_shape))),
                (slice(None) for _ in range(len(self.stack_shape))),
            )
            if stack_shape is not None
            else (slice(None) for _ in range(len(self.stack_shape)))
        )
        slices[self.shape_order[ShapeComponent.POINT]] = (
            itertools.chain(
                (np.newaxis for _ in range(len(point_shape) - len(self.point_shape))),
                (slice(None) for _ in range(len(self.point_shape))),
            )
            if point_shape is not None
            else (slice(None) for _ in range(len(self.point_shape)))
        )
        shapes[self.shape_order[ShapeComponent.BASIS]] = (
            self.basis_shape if basis_shape is None else basis_shape
        )
        shapes[self.shape_order[ShapeComponent.STACK]] = (
            self.stack_shape if stack_shape is None else stack_shape
        )
        shapes[self.shape_order[ShapeComponent.POINT]] = (
            self.point_shape if point_shape is None else point_shape
        )
        coefs = self.coefficients[*itertools.chain(*slices)]
        return Field(
            self.basis_shape if basis_shape is None else basis_shape,
            self.point_shape if point_shape is None else point_shape,
            (
                np_or_jnp(self).broadcast_to(coefs, shapes[0] + shapes[1] + shapes[2])
                if complete_broadcast
                else coefs
            ),
            shape_order=self.shape_order,
            use_jax=self.use_jax,
        )

    @staticmethod
    def are_broadcastable(*fields: "Field", strict_basis=True) -> bool:  # NOQA: ARG004
        """Two fields a and b are (fully) broadcastable if they are compatible and have
        broadcastable point shape. Since this relation is associative,
        more than two fields can be passed in.

        Args:
            fields (tuple[Field, ...]): The fields to broadcast.
            strict_basis (bool, optional): If true, the basis rule holds. Otherwise,
                only basis shape numpy-broadcastibility is checked.

        Returns:
            bool: True if the fields are broadcastable. False otherwise.
        """
        return Field.are_compatible(*fields) and _is_compatible(
            *(field.point_shape for field in fields)
        )

    @typing.overload
    @staticmethod
    def broadcast_fields_full(
        *fields: "Field", strict_basis=True, shapes_only: Literal[True]
    ) -> tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]: ...
    @typing.overload
    @staticmethod
    def broadcast_fields_full(
        *fields: "Field", strict_basis=True, shapes_only: Literal[False] = False
    ) -> tuple["Field", ...]: ...
    @staticmethod
    def broadcast_fields_full(
        *fields: "Field",
        strict_basis=True,  # NOQA: ARG004
        shapes_only: bool = False,
    ) -> tuple["Field", ...] | tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]:
        """Two fields a and b are (fully) broadcastable if they are compatible and have
        broadcastable point shape. Since this relation is associative,
        more than two fields can be passed in.

        Args:
            fields (tuple[Field, ...]): The fields to broadcast.
            strict_basis (bool, optional): If true, the basis rule holds. Otherwise,
                only basis shape numpy-broadcastibility is checked.
            shapes_only (bool, optional): If true, the target shape is returned and no
                array brouadcasting occurs.
        Raises:
            FieldShapeError: If the fields are not fully broadcastable together.

        Returns:
            tuple[Field, ...]: The broadcasted fields, in the order they were given.
        """
        if Field.are_broadcastable(*fields):
            basis_shape = np.broadcast_shapes(*[field.basis_shape for field in fields])
            stack_shape = np.broadcast_shapes(*[field.stack_shape for field in fields])
            point_shape = np.broadcast_shapes(*[field.point_shape for field in fields])
            if shapes_only:
                return (stack_shape, basis_shape, point_shape)
            return tuple(
                field.broadcast_to_shape(stack_shape, basis_shape, point_shape)
                for field in fields
            )

        message = "Cannot broadcast fields with incompatible shapes."
        raise FieldShapeError(message)

    @staticmethod
    def are_compatible(*fields: "Field", strict_basis=True) -> bool:
        """Two fields a and b are compatible if they have compatible bases
        (basis_shape equal or at least one of them is size 1 representing a constant)
        and they have broadcastable stack_shapes. This function checks them. Since
        this relation is associative, more than two fields can be passed in.

        Args:
            fields (tuple[Field, ...]): The fields to query compatibility.
            strict_basis (bool, optional): If true, the basis rule holds. Otherwise,
                only basis shape numpy-broadcastibility is checked.

        Returns:
            bool: True if the fields are compatible. False otherwise.
        """
        if strict_basis:
            return all(
                map(
                    lambda x: x[1],  # accumulator -> did nonempty tuple change?
                    itertools.accumulate(
                        (field.basis_shape for field in fields),
                        func=lambda a, b: (
                            a[0] if np.prod(b, dtype=int) == 1 else b,  # nonempty tuple
                            (np.prod(a[0], dtype=int) == 1)
                            or (np.prod(b, dtype=int) == 1)
                            or a[0] == b,  # if nonempty, did shape change?
                        ),
                        initial=((), True),
                    ),
                )
            ) and _is_compatible(*(field.stack_shape for field in fields))
        return _is_compatible(
            *(field.basis_shape for field in fields)
        ) and _is_compatible(*(field.stack_shape for field in fields))

    @staticmethod
    def broadcast_field_compatibility(
        *fields: "Field", strict_basis=True
    ) -> tuple["Field", ...]:
        """Two fields a and b are compatible if they have compatible bases
        (basis_shape equal or at least one of them is size 1 representing a constant)
        and they have broadcastable stack_shapes. Since
        this relation is associative, more than two fields can be passed in.

        This function broadcasts the fields to have the same stack and basis shapes if
        they are compatible, or raises an error if they are not.

        Args:
            fields (tuple[Field, ...]): The fields to broadcast.
            strict_basis (bool, optional): If true, the basis rule holds. Otherwise,
                only basis shape numpy-broadcastibility is checked.

        Raises:
            FieldShapeError: if the given fields are not all compatible.

        Returns:
            tuple[Field, ...]: The broadcasted fields, in the order they were given.
        """
        if Field.are_compatible(*fields, strict_basis=strict_basis):
            basis_shape = np.broadcast_shapes(*[field.basis_shape for field in fields])
            stack_shape = np.broadcast_shapes(*[field.stack_shape for field in fields])
            return tuple(
                field.broadcast_to_shape(stack_shape, basis_shape, field.point_shape)
                for field in fields
            )

        message = "Cannot broadcast fields with incompatible shapes."
        raise FieldShapeError(message)

    def _axis_field_to_numpy(
        self, index: FieldAxisIndexType, out_of_bounds_check: bool = True
    ) -> int:
        """Recovers the axis index (in terms of numpy) from the given field axis specifier.

        Args:
            index (FieldAxisIndexType): The field index to recover the axis of.
            out_of_bounds_check (bool, optional): Whether or not an out-of-bounds check
            occurs. As in python indexing, it is expected for
            -len(component_shape) <= index.index < len(component_shape), where
            component_shape is the shape of index.component. If out-of-bounds check is
            true, then an IndexError is thrown. Defaults to True.

        Raises:
            IndexError: if out_of_bounds_check is True, and the index is out of bounds.

        Returns:
            _type_: the integer index, in the numpy context.
        """
        comp: ShapeComponent = index[0]
        ind: int = index[1]
        shape = self.get_shape(comp)

        if ind < 0:
            if out_of_bounds_check and ind < -len(shape):
                message = (
                    f"Attempting to access axis {ind} ({-1-ind}) of shape {shape}."
                )
                raise IndexError(message)
            ind = len(shape) + ind
        elif out_of_bounds_check and ind >= len(shape):
            message = f"Attempting to access axis {ind} of shape {shape}."
            raise IndexError(message)

        return ind + self._component_offset(comp)

    def __eq__(self, other) -> bool:
        if not Field.are_broadcastable(self, other):
            return False

        return np_or_jnp(self, other).array_equiv(
            *(f.coefficients for f in Field.broadcast_fields_full(self, other))
        )



inv_mass_matrix = fnp.linalg.inv(
    fnp.moveaxis(
        mesh.mass_matrix(), (ShapeComponent.STACK, 0), (ShapeComponent.BASIS, 1)
    ),
    ShapeComponent.BASIS,
)


def solve_accel(wavefield):
    return -fnp.einsum(
        ShapeComponent.BASIS,
        "ij,j->i",
        inv_mass_matrix,
        mesh.integrate_grad_basis_dot_grad_field(wavefield, jacobian_scale=c2),
    )


Uddot = solve_accel(U)
while t < tmax:
    # use Newmark-beta
    U += dt * Udot + (0.5 * dt**2) * Uddot
    Udot += 0.5 * dt * Uddot
    Uddot = solve_accel(U)
    Udot += 0.5 * dt * Uddot
    t += dt



"""
The `fastfem.mesh` package contains utilities for generating and reading meshes.
"""

from .fundamentals import (
    RectangleMesh,
    SquareMesh,
    create_a_rectangle_mesh,
    create_a_square_mesh,
)
from .generator import (
    Domain,
    Line,
    Mesh,
    OneDElementType,
    Point,
    Submesh,
    Surface,
    TwoDElementType,
    ZeroDElementType,
    mesh,
)

__all__ = [
    "Domain",
    "Line",
    "Mesh",
    "OneDElementType",
    "Point",
    "RectangleMesh",
    "SquareMesh",
    "Submesh",
    "Surface",
    "TwoDElementType",
    "ZeroDElementType",
    "create_a_rectangle_mesh",
    "create_a_square_mesh",
    "mesh",
]



from . import numpy_similes_linalg as linalg
from .field import (
    Field,
    FieldAxisIndex,
    FieldConstructionError,
    FieldShapeError,
    ShapeComponent,
)
from .numpy_similes import abs, einsum, moveaxis, reshape, sum

__all__ = [
    "Field",
    "FieldAxisIndex",
    "FieldConstructionError",
    "FieldShapeError",
    "ShapeComponent",
    "abs",
    "einsum",
    "linalg",
    "moveaxis",
    "reshape",
    "sum",
]



import pathlib
import sys
from unittest.mock import MagicMock

import numpy as np
import pytest
import pyvista as pv

import fastfem.mesh as m
import fastfem.plotter as p

# Relevant constants for mesh building/videos
NODES_IN_HORIZONTAL_DIRECTION = 10
NODES_IN_VERTICAL_DIRECTION = 10
TOTAL_TIME = 1
FPS = 10
TIME_STEPS = int(TOTAL_TIME * FPS)


@pytest.fixture(params=["triangle", "quadrangle"])
def mesh(request: pytest.FixtureRequest) -> m.Mesh:
    """
    Fixture to create a mesh with different element types.

    Args:
        request: Pytest fixture.

    Returns:
        m.Mesh: The mesh object.
    """
    return m.create_a_rectangle_mesh(
        horizontal_length=1,
        vertical_length=1,
        nodes_in_horizontal_direction=NODES_IN_HORIZONTAL_DIRECTION,
        nodes_in_vertical_direction=NODES_IN_VERTICAL_DIRECTION,
        element_type=request.param,
    )


@pytest.fixture(params=[1, 2, 3])
def dummy_data() -> np.ndarray:
    """
    Fixture to create dummy data for testing.

    Returns:
        np.ndarray: Data.
    """
    left_temp = np.random.randint(0, 10)
    right_temp = np.random.randint(0, 10)
    top_temp = np.random.randint(0, 10)
    bottom_temp = np.random.randint(0, 10)
    min_temp = np.random.randint(10, 20)
    max_temp = np.random.randint(20, 30)
    data = np.random.uniform(
        low=min_temp,
        high=max_temp,
        size=(TIME_STEPS, NODES_IN_VERTICAL_DIRECTION, NODES_IN_HORIZONTAL_DIRECTION),
    )
    for i in range(TIME_STEPS):
        data[i, 0, :] = bottom_temp
        data[i, -1, :] = top_temp
        data[i, :, 0] = left_temp
        data[i, :, -1] = right_temp
    return data


def test_define_plotter(mesh: m.Mesh) -> None:
    """
    Tests if the VisualMesh class defines a mesh properly.

    Args:
        mesh: The mesh object.
    """
    visualizer = p.VisualMesh(mesh)
    grid = visualizer.define_plotter()
    assert isinstance(grid, pv.UnstructuredGrid)
    assert grid.n_cells > 0
    assert grid.n_points == mesh.number_of_nodes


@pytest.mark.parametrize("point_label", [True, False])
def test_plot_mesh(
    monkeypatch: pytest.MonkeyPatch,
    mesh: m.Mesh,
    point_label: bool,
) -> None:
    """
    Tests if the mesh is plotted properly, without data.

    Args:
        monkeypatch: Pytest fixture.
        mesh: The mesh object.
        point_label: Boolean value to determine whether the points are labeled.
        color: The color of the mesh/edges.
        edge_thickness: Thickness of the edges.
    """
    visualizer = p.VisualMesh(mesh)
    monkeypatch.setattr(pv.Plotter, "show", MagicMock())
    visualizer.plot_mesh(
        point_label=point_label,
    )


def test_plot_data(
    monkeypatch: pytest.MonkeyPatch, mesh: m.Mesh, dummy_data: np.ndarray
) -> None:
    """
    Tests if the mesh is plotted properly, for a single frame.

    Args:
        monkeypatch: Pytest fixture.
        mesh: The mesh object.
        dummy_data: The temperature data for each node, contained in a 2D array.
    """
    visualizer = p.VisualMesh(mesh)
    monkeypatch.setattr(pv.Plotter, "show", MagicMock())
    visualizer.plot_data(
        data=dummy_data[0],
    )


def test_animate_data(
    monkeypatch: pytest.MonkeyPatch,
    mesh: m.Mesh,
    dummy_data: np.ndarray,
) -> None:
    """
    Tests if the mesh is animated properly/

    Args:
        monkeypatch: Pytest fixture.
        mesh: The mesh object.
        dummy_data: The temperature data for each node, contained in a 3D array.
    """
    visualizer = p.VisualMesh(mesh)
    monkeypatch.setattr(pv.Plotter, "show", MagicMock())
    visualizer.animate_data(
        fps=FPS,
        total_time=TOTAL_TIME,
        data=dummy_data,
    )
    with pytest.raises(ValueError):
        visualizer.animate_data(
            fps=np.random.uniform(25.1, 100),
            total_time=TOTAL_TIME,
            data=dummy_data,
        )


@pytest.mark.skip(reason="Not currently supported")
def test_make_movie(
    tmp_path: pathlib.Path,
    mesh: m.Mesh,
    dummy_data: np.ndarray,
) -> None:
    """
    Tests if the movie is created.

    Args:
        monkeypatch: Pytest fixture.
        mesh: The mesh object.
        dummy_data: The temperature data for each node, contained in a 3D array.
    """
    visualizer = p.VisualMesh(mesh)
    filename = tmp_path / "test.mp4"
    visualizer.make_movie(
        filename=str(filename),
        fps=FPS,
        total_time=TOTAL_TIME,
        data=dummy_data,
    )
    assert filename.exists()


@pytest.mark.skipif(
    sys.platform in ["win32", "linux"],
    reason="Currently not supported on Windows/Linux",
)
def test_make_gif(
    tmp_path: pathlib.Path,
    mesh: m.Mesh,
    dummy_data: np.ndarray,
) -> None:
    """
    Tests if the gif is created.

    Args:
        monkeypatch: Pytest fixture.
        mesh: The mesh object.
        dummy_data: The temperature data for each node, contained in a 3D array
    """
    visualizer = p.VisualMesh(mesh)
    filename = tmp_path / "test.gif"
    visualizer.make_gif(
        filename=str(filename),
        fps=FPS,
        total_time=TOTAL_TIME,
        data=dummy_data,
    )
    assert filename.exists()



"""
The `fastfem.mesh.generator` module is a wrapper around Gmsh to create meshes by
defining points with coordinates, lines with points, and surfaces with lines. The
wrapper is designed to be used in a more object-oriented way to create meshes easily. It
also avoids creating duplicate entities in the Gmsh model by checking if the entity
already exists.
"""

import copy
import dataclasses
import pathlib
from typing import ClassVar, Literal, Optional

import gmsh
import numpy as np

__all__ = [
    "Domain",
    "Line",
    "Mesh",
    "OneDElementType",
    "Point",
    "Submesh",
    "Surface",
    "TwoDElementType",
    "ZeroDElementType",
    "mesh",
]


def __dir__() -> list[str]:
    return __all__


# Gmsh needs to be initialized before creating any entities:
gmsh.initialize()

# The available 2D element types:
ZeroDElementType = Literal["point"]
OneDElementType = Literal["line"]
TwoDElementType = Literal["triangle", "quadrangle"]

gmsh_element_type_dictionary: dict[
    Literal[1, 2, 3, 15], ZeroDElementType | OneDElementType | TwoDElementType
] = {
    1: "line",
    2: "triangle",
    3: "quadrangle",
    15: "point",
}

number_of_nodes_per_element: dict[
    ZeroDElementType | OneDElementType | TwoDElementType, int
] = {
    "point": 1,
    "line": 2,
    "triangle": 3,
    "quadrangle": 4,
}


@dataclasses.dataclass(kw_only=True, frozen=True)
class Submesh:
    type: ZeroDElementType | OneDElementType | TwoDElementType
    nodes: dict[int, np.ndarray]
    elements: dict[int, np.ndarray]

    @property
    def number_of_elements(self) -> int:
        """Return the number of elements."""
        return len(self.elements)

    @property
    def number_of_nodes_per_element(self) -> int:
        """Return the number of nodes per element."""
        return number_of_nodes_per_element[self.type]


@dataclasses.dataclass(kw_only=True, frozen=True)
class Domain:
    name: str
    tag: int
    dimension: int
    mesh: list[Submesh]


@dataclasses.dataclass(kw_only=True, frozen=True)
class Mesh:
    domains: list[Domain]

    @property
    def nodes(self) -> dict[int, np.ndarray]:
        """Return all the nodes."""
        nodes = {}
        for domain in self.domains:
            for submesh in domain.mesh:
                nodes.update(submesh.nodes)
        return nodes

    @property
    def number_of_nodes(self) -> int:
        """Return the number of nodes."""
        return len(self.nodes)

    def __getitem__(self, key: str) -> Domain:
        """Return the mesh of the domain with the given name.

        Args:
            key: The name of the domain.

        Returns:
            List of submeshes of the domain. Submeshes are required because each domain
            can have multiple element types.
        """
        for domain in self.domains:
            if domain.name == key:
                return domain

        message = f"Domain with the name {key} does not exist."
        raise KeyError(message)

    def __contains__(self, key: str) -> bool:
        """Check if the mesh has a domain with the given name.

        Args:
            key: The name of the domain.

        Returns:
            True if the domain exists, otherwise False.
        """
        return any(domain.name == key for domain in self.domains)

    def __iter__(self):
        """Return an iterator over the domains."""
        return iter(self.domains)


class Geometry:
    """This is a singleton class to store all the created geometric entities. It keeps
    track of all the points, lines, and surfaces created. If the user tries to create a
    new entity that already exists, this class will avoid the duplication of the entity,
    which is not done by Gmsh.

    Attributes:
        points: All the created points
        lines: All the created lines
        surfaces: All the created surfaces

    Methods:
        clear: Clear all the geometric entities.
        mesh: Generate the mesh using Gmsh.
    """

    _instance = None

    _points: ClassVar[list["Point"]] = []
    _lines: ClassVar[list["Line"]] = []
    _surfaces: ClassVar[list["Surface"]] = []

    _domains: ClassVar[dict[tuple[str, int], int]] = {}  # {(domain_name, dim): tag}

    _points_coordinates: np.ndarray = np.empty((0, 3), dtype=np.float64)
    _line_point_tags: np.ndarray = np.empty((0, 2), dtype=np.int64)
    _surface_line_tags: ClassVar[list[np.ndarray]] = []

    def __new__(cls):
        """If an instance of the class already exists, return it. Otherwise, create a
        new instance and return it. This is the Singleton design pattern."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def _does_it_already_exist(
        cls, entity: "Point | Line | Surface"
    ) -> Optional["Point | Line | Surface"]:
        """Check if the entity already exists. If it does, return the existing entity.

        Args:
            entity: The entity to check if it already exists.

        Returns:
            The existing entity if it already exists, otherwise None.
        """
        if isinstance(entity, Point):
            comparison = np.isclose(cls._points_coordinates, entity.coordinates).all(
                axis=1
            )
            entities = cls._points
        elif isinstance(entity, Line):
            comparison = (
                cls._line_point_tags == np.array([point.tag for point in entity.points])
            ).all(axis=1)
            entities = cls._lines

            # Check lines in the opposite direction:
            if not comparison.any():
                comparison = (
                    cls._line_point_tags
                    == np.array([point.tag for point in entity.points[::-1]])
                ).all(axis=1)
                if comparison.any():
                    where_true = np.where(comparison)[0][0]
                    new_line = copy.deepcopy(entities[where_true])
                    return -new_line

        elif isinstance(entity, Surface):
            entity_lines = np.array([line.tag for line in entity.outer_lines])
            comparison = np.array(
                [
                    np.array_equal(line_tag, entity_lines)
                    for line_tag in cls._surface_line_tags
                ]
            )
            entities = cls._surfaces

        if comparison.any():
            where_true = np.where(comparison)[0][0]
            return entities[where_true]

        return None

    @classmethod
    def _add(
        cls, entity: "Point | Line | Surface"
    ) -> Optional["Point | Line | Surface"]:
        """Add a geometric entity. If the entity already exists, return the existing
        entity. Otherwise, add the entity and return None.

        Args:
            entity: The entity to be added

        Returns:
            The existing entity if it already exists, otherwise the new entity
        """
        existing_entity = cls._does_it_already_exist(entity)
        if existing_entity:
            return existing_entity

        if isinstance(entity, Point):
            cls._points_coordinates = np.vstack(
                [cls._points_coordinates, entity.coordinates]
            )
            cls._points.append(entity)
        elif isinstance(entity, Line):
            cls._line_point_tags = np.vstack(
                [
                    cls._line_point_tags,
                    [point.tag for point in entity.points],
                ]
            )
            cls._lines.append(entity)
        elif isinstance(entity, Surface):
            cls._surface_line_tags.append(
                np.array(
                    np.array([line.tag for line in entity.outer_lines]),
                )
            )
            cls._surfaces.append(entity)

        return None

    @classmethod
    def _add_point(cls, point: "Point") -> "Point":
        """Add a point.

        Args:
            point: The point to be added.

        Returns:
            The existing point if it already exists, otherwise None.
        """
        return cls._add(point)  # type: ignore

    @classmethod
    def _add_line(cls, line: "Line") -> "Line":
        """Add a line.

        Args:
            line: The line to be added.

        Returns:
            The existing line if it already exists, otherwise None.
        """
        return cls._add(line)  # type: ignore

    @classmethod
    def _add_surface(cls, surface: "Surface") -> "Surface":
        """Add a surface.

        Args:
            surface: The surface to be added.

        Returns:
            The existing surface if it already exists, otherwise None.
        """
        return cls._add(surface)  # type: ignore

    @classmethod
    def clear(cls) -> None:
        """Clear all the geometric entities."""
        cls._points = []
        cls._lines = []
        cls._surfaces = []

        cls._points_coordinates = np.empty((0, 3), dtype=np.float64)
        cls._line_point_tags = np.empty((0, 2), dtype=np.int64)
        cls._surface_line_tags = []

        gmsh.clear()

    @classmethod
    def create_domains(cls) -> None:
        """Create physical groups in Gmsh."""
        gmsh.model.occ.synchronize()

        # domains = {domain_name: {dim: [tags]}}
        domains: dict[str, dict[int, list[int]]] = {}

        def find_their_domains(
            entities: list[Point] | list[Line] | list[Surface], dim: int
        ) -> None:
            for entity in entities:
                if entity.domain_name:
                    if entity.domain_name not in domains:
                        domains[entity.domain_name] = {dim: []}
                    try:
                        domains[entity.domain_name][dim].append(entity.tag)
                    except KeyError as e:
                        message = (
                            "The same domain name cannot be used for entities of"
                            " different dimensions (a point and a line cannot have the"
                            " same domain name, for example)."
                        )
                        raise ValueError(message) from e

        find_their_domains(cls._points, 0)
        find_their_domains(cls._lines, 1)
        find_their_domains(cls._surfaces, 2)

        for domain_name, domain_entities in domains.items():
            dim = next(iter(domain_entities.keys()))
            tags = domain_entities[dim]
            domain_tag = gmsh.model.add_physical_group(dim, tags, name=domain_name)
            cls._domains[domain_name, dim] = domain_tag

    @classmethod
    def mesh(cls) -> Mesh:
        """Generate the mesh using Gmsh."""
        cls.create_domains()
        gmsh.model.mesh.generate()

        domains: list[Domain] = []
        for domain_name_and_dim, domain_tag in cls._domains.items():
            name = domain_name_and_dim[0]
            dim = domain_name_and_dim[1]
            nodes_from_gmsh = gmsh.model.mesh.get_nodes_for_physical_group(
                dim, domain_tag
            )
            tags_of_nodes = nodes_from_gmsh[0]
            coordinates_of_nodes = np.array(nodes_from_gmsh[1]).reshape(-1, 3)

            geometric_entities = gmsh.model.get_entities_for_physical_group(
                dim, domain_tag
            )
            types_and_elements = {}
            for entity_tag in geometric_entities:
                elements_from_gmsh = gmsh.model.mesh.get_elements(dim, entity_tag)
                for i, type in enumerate(elements_from_gmsh[0]):
                    element_type = gmsh_element_type_dictionary[type]
                    if element_type not in types_and_elements:
                        types_and_elements[element_type] = {
                            "element_tags": np.array(elements_from_gmsh[1][i]),
                            "nodes_of_elements": np.array(
                                elements_from_gmsh[2][i]
                            ).reshape(-1, number_of_nodes_per_element[element_type]),
                        }
                    else:
                        types_and_elements[element_type]["element_tags"] = np.vstack(
                            [
                                types_and_elements[element_type]["element_tags"],
                                elements_from_gmsh[1][i],
                            ]
                        )
                        types_and_elements[element_type]["nodes_of_elements"] = (
                            np.vstack(
                                [
                                    types_and_elements[element_type][
                                        "nodes_of_elements"
                                    ],
                                    elements_from_gmsh[2][i],
                                ]
                            )
                        )

            # For each element type, create a Mesh object and append it to the meshes:
            meshes: list[Submesh] = []
            nodes: dict[int, np.ndarray] = {
                int(tag): coordinates
                for tag, coordinates in zip(
                    tags_of_nodes, coordinates_of_nodes, strict=False
                )
            }
            for element_type, elements_and_nodes in types_and_elements.items():
                elements: dict[int, np.ndarray] = {
                    int(tag): tags_of_nodes
                    for tag, tags_of_nodes in zip(
                        elements_and_nodes["element_tags"],
                        elements_and_nodes["nodes_of_elements"],
                        strict=False,
                    )
                }
                meshes.append(
                    Submesh(
                        type=element_type,
                        nodes=nodes,
                        elements=elements,
                    )
                )

            domains.append(
                Domain(name=name, tag=domain_tag, mesh=meshes, dimension=dim)
            )

        return Mesh(domains=domains)

    @property
    def points(self) -> list["Point"]:
        """Return all the created points."""
        return self._points

    @property
    def lines(self) -> list["Line"]:
        """Return all the created lines."""
        return self._lines

    @property
    def surfaces(self) -> list["Surface"]:
        """Return all the created surfaces."""
        return self._surfaces


@dataclasses.dataclass
class Point:
    """Create a point with the given coordinates.

    Args:
        x: The x-coordinate of the point.
        y: The y-coordinate of the point.
        z: The z-coordinate of the point.
        domain_name: The name of the domain the point belongs to. Defaults to None.
    """

    x: float
    y: float
    z: float
    domain_name: Optional[str] = None

    def __post_init__(self):
        self.coordinates = (self.x, self.y, self.z)

        point = Geometry()._add_point(self)
        if point:
            # This point already exists, send the old point instead of creating a new
            # one
            self.__dict__.update(point.__dict__)
        else:
            self.tag: int = gmsh.model.occ.add_point(self.x, self.y, self.z)


@dataclasses.dataclass
class Line:
    """Create a line with the given points.

    Args:
        start_point (Point): The start point of the line.
        end_point (Point): The end point of the line.
        number_of_nodes (Optional[int], optional): The number of nodes on the line. If
            provided, the line will be transfinite. Defaults to None.
        domain_name (Optional[str], optional): The name of the domain the line belongs
            to. Defaults to None.
    """

    start_point: Point
    end_point: Point
    number_of_nodes: Optional[int] = None
    domain_name: Optional[str] = None

    def __post_init__(
        self,
    ):
        self.points = (self.start_point, self.end_point)

        line = Geometry()._add_line(self)
        if line:
            # This line already exists, send the old line instead of creating a new one
            self.__dict__.update(line.__dict__)
        else:
            self.tag: int = gmsh.model.occ.add_line(
                self.points[0].tag, self.points[1].tag
            )

            self.transfinite = False
            if self.number_of_nodes:
                self.transfinite = True
                gmsh.model.occ.synchronize()
                gmsh.model.mesh.set_transfinite_curve(self.tag, self.number_of_nodes)

    def __neg__(self) -> "Line":
        """Create a new line with the opposite orientation."""
        self.points = (self.points[1], self.points[0])
        self.tag = -self.tag

        return self


@dataclasses.dataclass
class Surface:
    """Create a surface with the given lines.

    Args:
        outer_lines: The lines that form the surface. The lines must be connected (each
            line's end point is the start point of the next line).
        inner_lines: The lines that form the holes in the surface. Defaults to None.
        transfinite: If True, the surface will be transfinite. If transfinite, all the
            lines' number_of_nodes argument must be provided, and the surface must have
            3 or 4 lines, and there should be no inner lines. Defaults to False.
        element_type: The type of element to uFse. Defaults to "triangle".
            domain_name: The name of the domain the surface belongs to. Defaults to
                None.
    """

    outer_lines: list[Line]
    inner_lines: Optional[list[Line]] = None
    transfinite: bool = False
    element_type: TwoDElementType = "triangle"
    domain_name: Optional[str] = None

    def __post_init__(self):
        # Make sure all the lines are transfinite if the surface is transfinite
        if self.transfinite:
            lines_are_all_transfinite = all(
                line.transfinite for line in self.outer_lines
            )
            if not lines_are_all_transfinite:
                message = (
                    "If you would like to create a transfinite surface, all the lines'"
                    " number_of_nodes argument must be provided."
                )
                raise ValueError(message)

            if len(self.outer_lines) not in [3, 4]:
                message = "All the transfinite surfaces must have 3 or 4 lines."
                raise ValueError(message)

        # Make sure the points of lines are connected (each line's end point is the
        # start point of the next line) and lines are in clockwise order:
        def order_lines(lines: list[Line]) -> list[Line]:
            ordered_lines: list[Line] = []
            current_point_tag = lines[0].points[0].tag
            for line in lines:
                if line.points[0].tag == current_point_tag:
                    ordered_lines.append(line)
                    current_point_tag = line.points[1].tag
                elif line.points[1].tag == current_point_tag:
                    # If the line is in the opposite direction, use negative tag
                    ordered_lines.append(-line)
                    current_point_tag = line.points[1].tag
                else:
                    message = (
                        "Lines are not properly connected. Make sure the lines are"
                        " ordered."
                    )
                    raise ValueError(message)

            return ordered_lines

        ordered_outer_lines = order_lines(self.outer_lines)
        # Check if the loop is closed:
        if ordered_outer_lines[0].points[0] != ordered_outer_lines[-1].points[1]:
            message = "Lines do not form a closed loop."
            raise ValueError(message)

        self.outer_lines = ordered_outer_lines

        if self.inner_lines:
            ordered_inner_lines = order_lines(self.inner_lines)
            # Check if the loop is closed:
            if ordered_inner_lines[0].points[0] != ordered_inner_lines[-1].points[1]:
                message = "Inner lines do not form a closed loop."
                raise ValueError(message)

            self.inner_lines = ordered_inner_lines

        surface = Geometry()._add_surface(self)
        if surface:
            # This surface already exists, send the old surface instead of creating a
            # new one
            if self.inner_lines != surface.inner_lines:
                message = (
                    "There already exists a surface with the same outer lines but"
                    " different inner lines. We cannot create both of them as it will"
                    " cause duplication. Also, we cannot remove the old one since they"
                    " are not the same."
                )
                raise ValueError(message)
            self.__dict__.update(surface.__dict__)
        else:
            curve_loop_tags = []
            # Create a line loop and a plane surface
            outer_lines_tags = [line.tag for line in self.outer_lines]
            curve_loop_tags.append(gmsh.model.occ.add_curve_loop(outer_lines_tags))

            if self.inner_lines:
                inner_lines_tags = [line.tag for line in self.inner_lines]
                curve_loop_tags.append(gmsh.model.occ.add_curve_loop(inner_lines_tags))

            self.tag = gmsh.model.occ.add_plane_surface(curve_loop_tags)

            if self.transfinite or self.element_type == "quadrangle":
                gmsh.model.occ.synchronize()

            if self.transfinite:
                gmsh.model.mesh.set_transfinite_surface(self.tag)

            if self.element_type == "quadrangle":
                gmsh.model.mesh.set_recombine(2, self.tag)


def mesh(file_name: Optional[pathlib.Path] = None) -> Mesh:
    """Create a mesh from all the created geometric entities so far and return it as a
    `Mesh` object. If a file name is provided, write the mesh to the file in the Gmsh
    format.

    Args:
        file_name: The name of the file to write the mesh to. For example, "mesh.msh".

    Returns:
        The mesh as a `Mesh` object.
    """
    mesh = Geometry().mesh()

    if file_name:
        gmsh.write(file_name)

    Geometry().clear()

    return mesh



import numpy as np

import fastfem.mesh as m
import fastfem.plotter as p

NODES_IN_HORIZONTAL_DRECTION = 20
NODES_IN_VERTICAL_DIRECTION = 20
HORIZONTAL_LENGTH = 2
VERTICAL_LENGTH = 1
mesh = m.create_a_rectangle_mesh(
    horizontal_length=HORIZONTAL_LENGTH,
    vertical_length=VERTICAL_LENGTH,
    nodes_in_horizontal_direction=NODES_IN_HORIZONTAL_DRECTION,
    nodes_in_vertical_direction=NODES_IN_VERTICAL_DIRECTION,
    element_type="triangle",
    file_name=None,
)

visualizer = p.VisualMesh(mesh)
visualizer.plot_mesh()

# Time
TOTAL_TIME = 3
FPS = 25
TIME_STEPS = int(TOTAL_TIME * FPS)
TEMPERATURES = np.zeros(
    (TIME_STEPS, NODES_IN_VERTICAL_DIRECTION, NODES_IN_HORIZONTAL_DRECTION)
)

# Dummy Data
LEFT_TEMP = 0
RIGHT_TEMP = 0
TOP_TEMP = 0
BOTTOM_TEMP = 0
MIN_TEMP = 0
MAX_TEMP = 50

TEMPERATURES = np.random.uniform(
    low=MIN_TEMP,
    high=MAX_TEMP,
    size=(TIME_STEPS, NODES_IN_VERTICAL_DIRECTION, NODES_IN_HORIZONTAL_DRECTION),
)

for i in range(TIME_STEPS):
    TEMPERATURES[i, 0, :] = BOTTOM_TEMP
    TEMPERATURES[i, -1, :] = TOP_TEMP
    TEMPERATURES[i, :, 0] = LEFT_TEMP
    TEMPERATURES[i, :, -1] = RIGHT_TEMP


# Visualize
visualizer.animate_data(FPS, TOTAL_TIME, TEMPERATURES)

filename_movie = "filename_movie.mp4"
visualizer.make_movie(filename_movie, FPS, TOTAL_TIME, TEMPERATURES)

filename_gif = "filename_gif.gif"
visualizer.make_gif(filename_gif, FPS, TOTAL_TIME, TEMPERATURES)
